#!/usr/bin/env python
# -*- coding: utf-8 -*-
import io
import os
import time
import zipfile
import matplotlib.image as mpimg
import urllib
import numpy as np


__author__ = "Randall Balestriero"


class cassava:
    """Plant images classification.

    The data consists of two folders, a training folder that contains 5
    subfolders that contain the respective images for the different 5 classes
    and a test folder containing test images.

    Participants are to train their models using the images in the training
    folder and provide a submission file like the sample provided which contains
    the image name exactly matching the image name in the test folder and the
    corresponding class prediction with labels corresponding to the disease
    categories, cmd, healthy, cgm, cbsd, cbb.

    Please cite this paper if you use the dataset for your project:
    https://arxiv.org/pdf/1908.02900.pdf

    """

    classes = ["cbb", "cmd", "cbsd", "cgm", "healthy"]

    @staticmethod
    def download(path):
        """
        Download the cassava dataset and store the result into the given
        path

        Parameters
        ----------

            path: str
                the path where the downloaded files will be stored. If the
                directory does not exist, it is created.
        """

        # Check if directory exists
        if not os.path.isdir(path + "cassava"):
            print("Creating cassava Directory")
            os.mkdir(path + "cassava")
        # Check if file exists
        if not os.path.exists(path + "cassava/cassavaleafdata.zip"):
            url = (
                "https://storage.googleapis.com/emcassavadata/" + "cassavaleafdata.zip"
            )
            urllib.request.urlretrieve(url, path + "cassava/cassavaleafdata.zip")

    @staticmethod
    def load(path=None):
        """
        Parameters
        ----------
            path: str (optional)
                default ($DATASET_PATH), the path to look for the data and
                where the data will be downloaded if not present

        Returns
        -------

            train_images: array

            train_labels: array

            valid_images: array

            valid_labels: array

            test_images: array

            test_labels: array

        """

        if path is None:
            path = os.environ["DATASET_PATH"]

        cassava.download(path)

        t0 = time.time()

        # Loading the file
        data = {"train": [[], []], "test": [[], []], "validation": [[], []]}

        f = zipfile.ZipFile(path + "cassava/cassavaleafdata.zip")
        for filename in f.namelist():
            if ".jpg" not in filename:
                continue
            setname, foldername = filename.split("/")[1:3]
            img = mpimg.imread(io.BytesIO(f.read(filename)), "jpg")
            data[setname][0].append(img)
            data[setname][1].append(cassava.classes.index(foldername))

        train_images = np.array(data["train"][0])
        test_images = np.array(data["test"][0])
        valid_images = np.array(data["validation"][0])

        train_labels = np.array(data["train"][1])
        test_labels = np.array(data["test"][1])
        valid_labels = np.array(data["validation"][1])

        print("Dataset cassava loaded in {0:.2f}s.".format(time.time() - t0))

        return (
            train_images,
            train_labels,
            valid_images,
            valid_labels,
            test_images,
            test_labels,
        )



import os
import gzip
import urllib.request
import numpy as np
import time
import zipfile
import io
from .wavfile import read as wav_read
from tqdm import tqdm
import soundfile as sf


class TUTacousticscences2017:
    """
    Acoustic Scene classification

    TUT Acoustic Scenes 2017 dataset consists of recordings from various
    acoustic scenes, all having distinct recording locations. For each
    recording location, 3-5 minute long audio recording was captured.
    The original recordings were then split into segments with a length of 10
    seconds. These audio segments are provided in individual files.

    Roughly 11Go of data.

    Acoustic scenes for the task (15):

        Bus - traveling by bus in the city (vehicle)
        Cafe / Restaurant - small cafe/restaurant (indoor)
        Car - driving or traveling as a passenger, in the city (vehicle)
        City center (outdoor)
        Forest path (outdoor)
        Grocery store - medium size grocery store (indoor)
        Home (indoor)
        Lakeside beach (outdoor)
        Library (indoor)
        Metro station (indoor)
        Office - multiple persons, typical work day (indoor)
        Residential area (outdoor)
        Train (traveling, vehicle)
        Tram (traveling, vehicle)
        Urban park (outdoor)
    """

    classes = [
        "metro_station",
        "office",
        "park",
        "residential_area",
        "train",
        "tram",
        "library",
        "home",
        "grocery_store",
        "forest_path",
        "city_center",
        "car",
        "cafe/restaurant",
        "bus",
        "beach",
    ]

    def download(path):

        if path is None:
            path = os.environ["DATASET_path"]

        # Load the dataset (download if necessary) and set
        # the class attributes.

        print("Downloading TUTacousticscences2017")
        t = time.time()

        if not os.path.isdir(path + "TUTacousticscences2017"):
            print("\tCreating TUTacousticscences2017 Directory")
            os.mkdir(path + "TUTacousticscences2017")

        # training set
        url = (
            "https://zenodo.org/record/400515/files/"
            + "TUT-acoustic-scenes-2017-development.audio.{}.zip"
        )
        for part in range(1, 11):
            filename = (
                path
                + "TUTacousticscences2017/"
                + "TUT-acoustic-scenes-2017-development.audio.{}.zip"
            )
            if os.path.exists(filename.format(part)):
                continue
            print("\tDownloading train part {}".format(part))
            urllib.request.urlretrieve(url.format(part), filename.format(part))

        # meta data
        url = (
            "https://zenodo.org/record/400515/files/"
            + "TUT-acoustic-scenes-2017-development.meta.zip"
        )
        filename = (
            path
            + "TUTacousticscences2017/"
            + "TUT-acoustic-scenes-2017-development.meta.zip"
        )
        if not os.path.exists(filename):
            urllib.request.urlretrieve(url, filename)

        # test set
        url = (
            "https://zenodo.org/record/1040168/files/"
            + "TUT-acoustic-scenes-2017-evaluation.audio.{}.zip"
        )
        for part in range(1, 5):
            filename = (
                path
                + "TUTacousticscences2017/"
                + "TUT-acoustic-scenes-2017-evaluation.audio.{}.zip"
            )
            if os.path.exists(filename.format(part)):
                continue
            print("\tDownloading test part {}".format(part))
            urllib.request.urlretrieve(url.format(part), filename.format(part))

        # meta data
        url = (
            "https://zenodo.org/record/1040168/files/"
            + "TUT-acoustic-scenes-2017-evaluation.meta.zip"
        )
        filename = (
            path
            + "TUTacousticscences2017/"
            + "TUT-acoustic-scenes-2017-evaluation.meta.zip"
        )
        if not os.path.exists(filename):
            urllib.request.urlretrieve(url, filename)

        t = time.time() - t
        print("TUT-acoustic-scenes-2017 downloaded in {} sec.".format(t))

    def load(path=None):
        if path is None:
            path = os.environ["DATASET_PATH"]

        TUTacousticscences2017.download(path)
        t = time.time()

        # meta data
        filename = (
            path
            + "TUTacousticscences2017/"
            + "TUT-acoustic-scenes-2017-development.meta.zip"
        )
        meta = zipfile.ZipFile(filename)
        meta_names = list()
        wav_names = list()
        wav_labels = list()
        for filename in meta.namelist():
            if "train" not in filename and "evaluate" not in filename:
                continue
            meta_names.append(filename.split("/")[-1][:-4])
            content = np.loadtxt(
                io.BytesIO(meta.read(filename)), delimiter="\t", dtype="str"
            )
            wav_names.append(list(content[:, 0]))
            wav_labels.append(list(content[:, 1]))
            for j in range(len(wav_names[-1])):
                wav_names[-1][j] = wav_names[-1][j].split("/")[-1]

        folds = list()
        wavs = list()
        labels = list()

        filename = (
            path
            + "TUTacousticscences2017/"
            + "TUT-acoustic-scenes-2017-development.audio.{}.zip"
        )

        for part in range(1, 11):
            # load wavs
            f = zipfile.ZipFile(filename.format(part))
            for name in tqdm(
                f.namelist(), ascii=True, desc="Train Part:{}/10".format(part)
            ):
                if ".wav" not in name:
                    continue
                wavfile = f.read(name)
                byt = io.BytesIO(wavfile)
                wavs.append(sf.read(byt)[0].astype("float32"))
                nn = name.split("/")[-1]
                add_label = 1
                folds.append([])
                for meta_name, wav_name, wav_label in zip(
                    meta_names, wav_names, wav_labels
                ):
                    if nn in wav_name:
                        if add_label:
                            index = wav_name.index(nn)
                            labels.append(wav_label[index])
                            add_label = 0
                        folds[-1].append(meta_name)

        # now format the folds
        train_folds = np.zeros((len(folds), 4), dtype="bool")
        for i in range(len(folds)):
            for fold in folds[i]:
                if "train" in fold:
                    train_folds[i, int(fold.split("_")[0][-1]) - 1] = 1

        # now deal with the test set

        # meta data
        filename = (
            path
            + "TUTacousticscences2017/"
            + "TUT-acoustic-scenes-2017-evaluation.meta.zip"
        )
        meta = zipfile.ZipFile(filename)
        for filename in meta.namelist():
            #            if 'map.txt' in filename:
            #                content = np.loadtxt(io.BytesIO(meta.read(filename)),
            #                                     delimiter='\t', dtype='str')
            if "evaluate.txt" in filename:
                targets = np.loadtxt(
                    io.BytesIO(meta.read(filename)), delimiter="\t", dtype="str"
                )
        #        content_0 = [c.split('/')[1] for c in content[:, 0]]
        #        content_1 = [c.split('/')[1] for c in content[:, 1]]
        targets_0 = [c.split("/")[1] for c in targets[:, 0]]
        targets_1 = list(targets[:, 1])

        #        for i in range(len(content_0)):
        #            content_1[i] = targets_1[targets_0.index(content_1[i])]
        mapping = dict(zip(targets_0, targets_1))

        test_wavs = list()
        test_labels = list()
        filename = (
            path
            + "TUTacousticscences2017/"
            + "TUT-acoustic-scenes-2017-evaluation.audio.{}.zip"
        )

        for part in range(1, 5):
            # load wavs
            f = zipfile.ZipFile(filename.format(part))
            for name in tqdm(
                f.namelist(), ascii=True, desc="Test Data Part{}/4".format(part)
            ):
                if ".wav" not in name:
                    continue
                byt = io.BytesIO(f.read(name))
                test_wavs.append(sf.read(byt)[0].astype("float32"))
                test_labels.append(mapping[name.split("/")[-1]])

        # turn the labels into integers
        mapping = dict(zip(TUTacousticscences2017.classes, list(range(15))))
        labels = [mapping[l] for l in labels]
        test_labels = [mapping[l] for l in test_labels]

        return (
            np.array(wavs),
            np.array(labels),
            np.array(test_wavs),
            np.array(test_labels),
            train_folds,
        )


# for i in range(1, 22):
#    filename = 'https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.{}.zip?download=1'.format(i)



import os
import pickle, gzip
import urllib.request
import numpy as np
import tarfile
import time
from tqdm import tqdm
from scipy.io.wavfile import read as wav_read


class gtzan:
    """music genre classification

    This dataset was used for the well known paper in genre classification
    "Musical genre classification of audio signals" by G. Tzanetakis
    and P. Cook in IEEE Transactions on Audio and Speech Processing 2002.

    Unfortunately the database was collected gradually and very early on in my
    research so I have no titles (and obviously no copyright permission etc).
    The files were collected in 2000-2001 from a variety of sources including
    personal CDs, radio, microphone recordings, in order to represent a variety
    of recording conditions. Nevetheless I have been providing it to researchers
    upon request mainly for comparison purposes etc. Please contact George
    Tzanetakis (gtzan@cs.uvic.ca) if you intend to publish experimental results
    using this dataset.

    There are some practical and conceptual issues with this dataset, described
    in "The GTZAN dataset: Its contents, its faults, their effects on
    evaluation, and its future use" by B. Sturm on arXiv 2013.
    """

    name2class = {
        "blues": 0,
        "classical": 1,
        "country": 2,
        "disco": 3,
        "hiphop": 4,
        "jazz": 5,
        "metal": 6,
        "pop": 7,
        "reggae": 8,
        "rock": 9,
    }

    def download(path=None):
        if path is None:
            path = os.environ["DATASET_path"]

        t0 = time.time()

        print("Downloading gtzan")

        # Check if directory exists
        if not os.path.isdir(path + "gtzan"):
            print("\tCreating Directory")
            os.mkdir(path + "gtzan")

        # Check if file exists
        if not os.path.exists(path + "gtzan//genres.tar.gz"):
            url = "http://opihi.cs.uvic.ca/sound/genres.tar.gz"
            urllib.request.urlretrieve(url, path + "gtzan/genres.tar.gz")

    def load(path=None):

        if path is None:
            path = os.environ["DATASET_PATH"]
        gtzan.download(path)

        t0 = time.time()

        print("Loading gtzan")

        tar = tarfile.open(path + "gtzan/genres.tar.gz", "r:gz")

        # Load train set
        train_songs = list()
        train_labels = list()
        names = tar.getnames()
        names = tar.getmembers()
        for name in tqdm(names, ascii=True, total=1000):
            if "wav" not in name.name:
                continue
            f = tar.extractfile(name.name)  # .read()
            train_songs.append(wav_read(f)[1])
            t = name.name.split("/")[1]
            train_labels.append(gtzan.name2class[t])

        N = np.min([len(w) for w in train_songs])
        train_songs = [w[:N] for w in train_songs]

        train_songs = np.stack(train_songs).astype("float32")
        train_labels = np.array(train_labels).astype("int32")

        print("Dataset gtzan loaded in{0:.2f}s.".format(time.time() - t0))

        return train_songs, train_labels



import os
import pickle, gzip
import urllib.request
import numpy as np
import time
import tarfile
from tqdm import tqdm
import zipfile
from scipy.io.wavfile import read as wav_read
import io


class FSDKaggle2018:
    """FSDKaggle2018 Sound Classification
    https://zenodo.org/record/2552860
    """

    def download(path):

        # Check if directory exists
        if not os.path.isdir(path + "FSDKaggle2018"):
            print("\tCreating FSDKaggle2018 Directory")
            os.mkdir(path + "FSDKaggle2018")

        url = "https://zenodo.org/record/2552860/files/FSDKaggle2018.audio_test.zip?download=1"
        # Check if file exists
        if not os.path.exists(path + "FSDKaggle2018/audio_test.zip"):
            print("\tDownloading test set")
            urllib.request.urlretrieve(url, path + "FSDKaggle2018/audio_test.zip")
        url = "https://zenodo.org/record/2552860/files/FSDKaggle2018.audio_train.zip?download=1"
        if not os.path.exists(path + "FSDKaggle2018/audio_train.zip"):
            print("\tDownloading train set")
            urllib.request.urlretrieve(url, path + "FSDKaggle2018/audio_train.zip")
        url = (
            "https://zenodo.org/record/2552860/files/FSDKaggle2018.meta.zip?download=1"
        )
        if not os.path.exists(path + "FSDKaggle2018/meta.zip"):
            print("\tDownloading meta set")
            urllib.request.urlretrieve(url, path + "FSDKaggle2018/meta.zip")

    def load(path=None):
        """
        Parameters
        ----------
        
        path: str (optional)
                default $DATASET_path), the path to look for the data and
                where the data will be downloaded if not present

        Returns
        -------

        wavs: array
            the wavs as a numpy array (matrix) with first dimension the data
            and second dimension time
        
        labels: array
            the labels of the final classes (41 different ones) as a integer
            vector
        """

        if path is None:
            path = os.environ["DATASET_PATH"]
        FSDKaggle2018.download(path)
        t0 = time.time()

        f = zipfile.ZipFile(path + "FSDKaggle2018/audio_train.zip")
        wavs_train = list()
        names_train = list()
        for filename in tqdm(f.namelist(), ascii=True, desc="Loading train set"):
            if ".wav" not in filename:
                continue
            wavfile = f.read(filename)
            byt = io.BytesIO(wavfile)
            wavs_train.append(wav_read(byt)[1].astype("float32"))
            names_train.append((filename.split("/")[-1]))

        f = zipfile.ZipFile(path + "FSDKaggle2018/audio_test.zip")
        wavs_test = list()
        names_test = list()
        for filename in tqdm(f.namelist(), ascii=True, desc="Loading test set"):
            if ".wav" not in filename:
                continue
            wavfile = f.read(filename)
            byt = io.BytesIO(wavfile)
            wavs_test.append(wav_read(byt)[1].astype("float32"))
            names_test.append((filename.split("/")[-1]))

        f = zipfile.ZipFile(path + "FSDKaggle2018/meta.zip")
        meta_train = np.loadtxt(
            io.BytesIO(f.read("FSDKaggle2018.meta/train_post_competition.csv")),
            delimiter=",",
            skiprows=1,
            dtype="str",
        )
        meta_test = np.loadtxt(
            io.BytesIO(
                f.read("FSDKaggle2018.meta/test_post_competition_scoring_clips.csv")
            ),
            delimiter=",",
            skiprows=1,
            dtype="str",
        )

        filenames = list(meta_train[:, 0])
        labels_train, verified, fsid_train = [], [], []
        for i in range(len(wavs_train)):
            index = filenames.index(names_train[i])
            labels_train.append(meta_train[index][1])
            verified.append(meta_train[index][2])
            fsid_train.append(meta_train[index][3])

        filenames = list(meta_test[:, 0])
        labels_test, usage, fsid_test = [], [], []
        for i in range(len(wavs_test)):
            index = filenames.index(names_test[i])
            labels_test.append(meta_test[index][1])
            usage.append(meta_test[index][2])
            fsid_test.append(meta_test[index][3])
        dataset = {
            "wavs_train": wavs_train,
            "labels_train": labels_train,
            "verified_train": verified,
            "fsid_train": fsid_train,
            "wavs_test": wavs_test,
            "labels_test": labels_test,
            "usage_test": usage,
            "fsid_test": fsid_test,
        }
        return dataset



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Randall Balestriero"

try:
    import pygraphviz as pgv
except ImportError:
    raise ImportError("Missing pygraphviz dependency.")

from . import tensor as T


def compute_graph(var, graph=None):
    if graph is None:
        graph = G = pgv.AGraph(strict=False, directed=True)
    graph.add_node(str(var))
    node = graph.get_node(var)
    node.attr["style"] = "filled"
    if type(var) == T.Op:
        node.attr["color"] = "green"
    elif type(var) == T.Variable:
        node.attr["shape"] = "rectangle"
        node.attr["color"] = "yellow"
    elif type(var) == T.Placeholder:
        node.attr["shape"] = "rectangle"
        node.attr["color"] = "blue"
    else:
        node.attr["shape"] = "rectangle"
        node.attr["color"] = "grey"

    if type(var) == T.Op:
        for v in var.args:
            try:
                graph.get_node(str(v))
            except KeyError:
                graph = compute_graph(v, graph)
                graph.add_edge(str(v), str(var))
        for v in var.kwargs:
            try:
                graph.get_node(str(var.kwargs[v]))
            except KeyError:
                graph = compute_graph(var.kwargs[v], graph)
                graph.add_edge(str(var.kwargs[v]), str(var))
    return graph



import gzip
import io
import os
import time
import urllib.request
import zipfile

import numpy as np
import tqdm


class emnist:
    """Grayscale digit/letter classification.

    The EMNIST Dataset

    Authors
    -------
    Gregory Cohen, Saeed Afshar, Jonathan Tapson,
    and Andre van Schaik
    
    The MARCS Institute for Brain, Behaviour and Development
    Western Sydney University
    Penrith, Australia 2751
    
    Email: g.cohen@westernsydney.edu.au
    
    What is it?
    -----------
    The EMNIST dataset is a set of handwritten character digits
    derived from the NIST Special Database 19
    (https://www.nist.gov/srd/nist-special-database-19) and
    converted to a 28x28 pixel image format and dataset structure
    that directly matches the MNIST dataset
    (http://yann.lecun.com/exdb/mnist/). Further information on
    the dataset contents and conversion process can be found in
    the paper available at https://arxiv.org/abs/1702.05373v1.
    
    Formats
    -------
    The dataset is provided in two file formats. Both versions of
    the dataset contain identical information, and are provided
    entirely for the sake of convenience. The first dataset is
    provided in a Matlab format that is accessible through both
    Matlab and Python (using the scipy.io.loadmat function). The
    second version of the dataset is provided in the same binary
    format as the original MNIST dataset as outlined in
    http://yann.lecun.com/exdb/mnist/
    
    Dataset Summary
    ---------------
    There are six different splits provided in this dataset.
    A short summary of the dataset is provided below:

    EMNIST ByClass:EMNIST814,255 characters. 62 unbalanced classes
    EMNIST ByMerge:     814,255 characters. 47 unbalanced classes
    EMNIST Balanced:Balanced131,600 characters. 47 balanced classes.
    EMNIST Letters:EMNIST145,600 characters. 26 balanced classes.
    EMNIST Digits:EMNIST280,000 characters. 10 balanced classes.
    EMNIST MNIST:EMNIST 70,000 characters. 10 balanced classes.

    The full complement of the NIST Special Database 19 is
    available in the ByClass and ByMerge splits. The EMNIST
    Balanced dataset contains a set of characters with an equal
    number of samples per class. The EMNIST Letters dataset
    merges a balanced set of the uppercase and lowercase letters
    into a single 26-class task. The EMNIST Digits and EMNIST
    MNIST dataset provide balanced handwritten digit datasets
    directly compatible with the original MNIST dataset.

    Please refer to the EMNIST paper (available at
    https://arxiv.org/abs/1702.05373v1) for further details of
    the dataset structure.

    How to cite
    -----------
    Please cite the following paper when using or referencing
    the dataset:

    Cohen, G., Afshar, S., Tapson, J., & van Schaik, A. (2017).
    EMNIST: an extension of MNIST to handwritten letters.
    Retrieved from http://arxiv.org/abs/1702.05373

    Files
    -----
    The dataset consists of the following files:

    .
    +-- gzip.zip
    ¦   +-- emnist-balanced-mapping.txt
    ¦   +-- emnist-balanced-test-images-idx3-ubyte.gz
    ¦   +-- emnist-balanced-test-labels-idx1-ubyte.gz
    ¦   +-- emnist-balanced-train-images-idx3-ubyte.gz
    ¦   +-- emnist-balanced-train-labels-idx1-ubyte.gz
    ¦   +-- emnist-byclass-mapping.txt
    ¦   +-- emnist-byclass-test-images-idx3-ubyte.gz
    ¦   +-- emnist-byclass-test-labels-idx1-ubyte.gz
    ¦   +-- emnist-byclass-train-images-idx3-ubyte.gz
    ¦   +-- emnist-byclass-train-labels-idx1-ubyte.gz
    ¦   +-- emnist-bymerge-mapping.txt
    ¦   +-- emnist-bymerge-test-images-idx3-ubyte.gz
    ¦   +-- emnist-bymerge-test-labels-idx1-ubyte.gz
    ¦   +-- emnist-bymerge-train-images-idx3-ubyte.gz
    ¦   +-- emnist-bymerge-train-labels-idx1-ubyte.gz
    ¦   +-- emnist-digits-mapping.txt
    ¦   +-- emnist-digits-test-images-idx3-ubyte.gz
    ¦   +-- emnist-digits-test-labels-idx1-ubyte.gz
    ¦   +-- emnist-digits-train-images-idx3-ubyte.gz
    ¦   +-- emnist-digits-train-labels-idx1-ubyte.gz
    ¦   +-- emnist-letters-mapping.txt
    ¦   +-- emnist-letters-test-images-idx3-ubyte.gz
    ¦   +-- emnist-letters-test-labels-idx1-ubyte.gz
    ¦   +-- emnist-letters-train-images-idx3-ubyte.gz
    ¦   +-- emnist-letters-train-labels-idx1-ubyte.gz
    ¦   +-- emnist-mnist-mapping.txt
    ¦   +-- emnist-mnist-test-images-idx3-ubyte.gz
    ¦   +-- emnist-mnist-test-labels-idx1-ubyte.gz
    ¦   +-- emnist-mnist-train-images-idx3-ubyte.gz
    ¦   +-- emnist-mnist-train-labels-idx1-ubyte.gz
    +-- matlab.zip
        +-- emnist-balanced.mat
            +-- emnist-byclass.mat
                +-- emnist-bymerge.mat
                    +-- emnist-digits.mat
                        +-- emnist-letters.mat
                            +-- emnist-mnist.mat
                            +-- Readme.txt

    """

    @staticmethod
    def download(path):
        """
        Download the EMNIST dataset and store the result into
        the given path

        Parameters
        ----------

            path: str
                the path where the downloaded files will be
                stored. If the directory does not exist, it is created.
        """

        # Check if directory exists
        if not os.path.isdir(path + "emnist"):
            print("Creating emnist Directory")
            os.mkdir(path + "emnist")

        # Check if file exists
        if not os.path.exists(path + "emnist/gzip.zip"):
            td = time.time()
            print("Downloading emnist")
            url = "http://www.itl.nist.gov/iaui/vip/cs_links/EMNIST/gzip.zip"
            urllib.request.urlretrieve(url, path + "emnist/gzip.zip")

    @staticmethod
    def _read_images(filename, folder):

        mat = folder.read(filename)
        f = gzip.open(io.BytesIO(mat), "rb")

        magic = f.read(4)
        magic = int.from_bytes(magic, "big")
        print("Magic is:", magic)

        # Number of images in next 4 bytes
        noimg = f.read(4)
        noimg = int.from_bytes(noimg, "big")

        # Number of rows in next 4 bytes
        norow = f.read(4)
        norow = int.from_bytes(norow, "big")

        # Number of columns in next 4 bytes
        nocol = f.read(4)
        nocol = int.from_bytes(nocol, "big")

        images = np.empty((noimg, norow, nocol), "float32")

        for i in tqdm.tqdm(range(noimg), ascii=True):
            for r in range(norow):
                for c in range(nocol):
                    images[i, c, r] = float(int.from_bytes(f.read(1), "big"))

        f.close()

        return images

    @staticmethod
    def _read_labels(filename, folder):

        mat = folder.read(filename)
        f = gzip.open(io.BytesIO(mat), "rb")

        magic = f.read(4)
        magic = int.from_bytes(magic, "big")
        print("Magic is:", magic)

        # Number of images in next 4 bytes
        nolab = f.read(4)
        nolab = int.from_bytes(nolab, "big")

        labels = [f.read(1) for i in range(nolab)]
        labels = [int.from_bytes(label, "big") for label in labels]

        f.close()

        return np.array(labels)

    @staticmethod
    def load(option="byclass", path=None):
        """
        Parameters
        ----------

            option: str
                the emnist dataset to load (byclass, bymerge, balanced, letters,
                digits, mnist)

            path: str (optional)
                default ($DATASET_PATH), the path to look for the data and
                where the data will be downloaded if not present

        Returns
        -------

            train_images: array

            train_labels: array

            valid_images: array

            valid_labels: array

            test_images: array

            test_labels: array

        """

        if path is None:
            path = os.environ["DATASET_PATH"]

        emnist.download(path)

        t0 = time.time()

        # Loading the file
        print("Loading emnist")
        folder = zipfile.ZipFile(path + "emnist/gzip.zip")
        filename = "gzip/emnist-{}-{}-{}-idx{}-ubyte.gz"
        x_test = emnist._read_images(
            filename.format(option, "test", "images", 3), folder
        )
        x_train = emnist._read_images(
            filename.format(option, "train", "images", 3), folder
        )
        y_test = emnist._read_labels(
            filename.format(option, "test", "labels", 1), folder
        )
        y_train = emnist._read_labels(
            filename.format(option, "train", "labels", 1), folder
        )

        return x_train, y_train, x_test, y_test



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__all__ = ["layers", "initializers", "losses", "schedules", "optimizers"]


from .ops_nn import *
from . import layers
from . import losses
from . import schedules
from . import optimizers
from . import initializers



import symjax.tensor as T
import symjax as sj
import jax
import networkx as nx
import matplotlib.pyplot as plt


bin_rules = [
    [
        lambda *args: args[0] == 0 or args[1] == 0,
        jax.numpy.add,
        lambda *args: args[0] if args[1] == 0 else args[0],
    ],
    [lambda *args: len(args) == 1, jax.numpy.add, lambda *args: 2 * args[0]],
    [lambda *args: args[1] == 1, jax.numpy.true_divide, lambda *args: args[0]],
    [
        lambda *args: len(args) == 1,
        jax.numpy.true_divide,
        lambda *args: T.ones(args[0].shape, args[0].dtype),
    ],
    [
        lambda *args: args[0] == 1 or args[1] == 1,
        jax.numpy.multiply,
        lambda *args: args[0] if args[1] == 1 else args[1],
    ],
]


def simplify_add(graph):
    to_search = list(graph.nodes.keys())
    while len(to_search):
        j = to_search[-1]
        if type(j) == T.Op:

            for rule in bin_rules:

                if graph.get_node_attribute(j, "jax_function") == rule[1]:

                    predecessors = list(graph.predecessors(j))
                    try:
                        values = rule[0](*predecessors)
                    except IndexError:
                        values = False
                    if not values:
                        continue
                    new_node = rule[2](*predecessors)

                    for s in graph.successors(j):
                        graph.add_edge(new_node, s)
                        graph[new_node][s]["name"] = graph[j][s]["name"]

                        graph.remove_node(j)
                    break

        to_search.pop(-1)


a = T.ones(10)
b = a + 0
c = b / 1
d = c * 2

graph = sj.current_graph()

plt.subplot(211)
nx.draw(
    graph,
    with_labels=True,
    node_size=600,
    alpha=0.4,
    node_shape="s",
    pos=nx.spring_layout(graph),
)

simplify_add(sj.current_graph())

plt.subplot(212)
nx.draw(
    graph,
    with_labels=True,
    node_size=600,
    alpha=0.4,
    node_shape="s",
    pos=nx.spring_layout(graph),
)

f = sj.function(outputs=d)

print(f(), f())
plt.axis("off")
plt.show()



import os
from tqdm import tqdm
import urllib.request
import numpy as np
import time
import io
import imageio
import tarfile
from scipy.ndimage import imread


class cub200:
    """ Image classification of bird species.
    The `CUB-200 <http://www.vision.caltech.edu/visipedia/CUB-200-2011.html>`_.
    dataset  contains  11,788  images of  200  bird  species
    This dataset contains 8732 labeled sound excerpts (<=4s) of urban sounds 
    from 10 classes: air_conditioner, car_horn, children_playing, 
    dog_bark, drilling, enginge_idling, gun_shot, jackhammer, siren, 
    and street_music. The classes are drawn from the 
    `urban sound taxonomy <https://urbansounddataset.weebly.com/taxonomy.html>`_. 
    The dataset is obtained from `Kaggle <https://www.kaggle.com/pavansanagapati/urban-sound-classification>_`
    """

    def __init__(self, target="class", data_format="NCHW"):
        """set up the options
        target = {class,bounding_box}
        crop-images True or False (whether to use the cropped images)
        """

    data_format = "NCHW"
    """(default)"""
    datum_shape = (3, None, None)
    """(default)"""
    given_test_set = True
    """"""
    given_valid_set = False
    """"""
    given_unlabeled = False
    """"""
    n_classes = 200
    """"""
    # To be defined during loading
    classes = {}
    name = "cub200"

    def load(data_format="NCHW"):
        """Load the dataset (download if necessary) and adapt
        the class attributes based on the given data format.

        :param data_format: (optional, default 'NCHW'), if different than default, adapts :mod:`data_format` and :mod:`datum_shape`
        :type data_format: 'NCHW' or 'NHWC'
        :return: return the train and test set, each as a couple (images,labels)
        :rtype: [(train_images,train_labels),
                (test_images,test_labels)]
        """
        t = time.time()

        PATH = os.environ["DATASET_PATH"]

        if not os.path.isdir(PATH + "cub200"):
            print("Creating Directory")
            os.mkdir(PATH + "cub200")

        if not os.path.exists(PATH + "cub200/CUB_200_2011.tgz"):
            td = time.time()
            print("\tDownloading Data")
            url = "http://www.vision.caltech.edu/visipedia-data/CUB-200-2011/CUB_200_2011.tgz"
            urllib.request.urlretrieve(url, PATH + "cub200/CUB_200_2011.tgz")
            print("\tDone in {:.2f}".format(time.time() - td))

        print("Loading caltechbird")
        tar = tarfile.open(PATH + "cub200/CUB_200_2011.tgz", "r:gz")

        # Load the class names
        f = tar.extractfile("CUB_200_2011/classes.txt")
        names = np.loadtxt(f, dtype="str")
        cub200.classes = dict([[c, n.split(".")[1]] for c, n in enumerate(names[:, 1])])

        # Load Bounding boxes
        f = tar.extractfile("CUB_200_2011/bounding_boxes.txt")
        boxes = np.loadtxt(f, dtype="int32")
        bounding_boxes = dict()
        for i in range(boxes.shape[0]):
            bounding_boxes[str(boxes[i, 0])] = boxes[i, 1:]

        # Load dataset
        labels = list()
        boxes = list()
        data = np.empty((11788, 3, 500, 500)) * np.nan
        cpt = 0
        nb_removed_images = 0
        for member in tqdm(tar.getmembers()):
            if "CUB_200_2011/images/" in member.name and "jpg" in member.name:
                class_ = member.name.split("/")[2].split(".")[0]
                image_id = member.name.split("_")[-1][:-4]
                f = tar.extractfile(member)
                im = imageio.imread(f, format="jpg")
                if len(im.shape) == 2:
                    nb_removed_images += 1
                    print("i")
                    continue
                data[cpt, :, : im.shape[0], : im.shape[1]] = im.transpose([2, 0, 1])
                labels.append(int(class_))
                #                boxes.append(bounding_boxes[image_id])
                cpt += 1
        if nb_removed_images:
            data = data[:-nb_removed_images]
        labels = np.array(labels).astype("int32")
        if data_format == "NHWC":
            cub200.datum_shape = (None, None, 3)
            cub200.data_format = data_format
            data = data.transpose([0, 2, 3, 1])

        print("Dataset cub200 loaded in", "{0:.2f}".format(time.time() - t), "s.")

        return data, labels, np.asarray(boxes)



import re
from functools import wraps

import warnings
import jax
import jax.numpy as jnp
import numpy

import symjax


def list2tuple(item):
    if type(item) == list or type(item) == tuple and type(item) != slice:
        return tuple([list2tuple(i) for i in item])
    else:
        return item


def _add_method(cls):
    # important we keep the self inside the function call !
    def decorator(func, name=""):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            return func(self, *args, **kwargs)

        if name == "":
            setattr(cls, func.__name__, wrapper)
        else:
            setattr(cls, name, wrapper)
        return func  # returning func means func can still be used normally

    return decorator


def _args_formatting(args, extra_args, indices):
    """ utility function to be used in the Tensor class to correctly join the
    args and extra_args based on the indices

    Parameters:
    -----------

    args: List

    extra_args: List

    indices: List of binary values
        the indices (one per element) to join args and extra_args in the correct
        order

    """
    output = ()
    arg_iterator = iter(args)
    extra_arg_iterator = iter(extra_args)
    for i in indices:
        if i:
            output += (next(extra_arg_iterator),)
        else:
            output += (next(arg_iterator),)
    return output


def isvar(item):
    """ check whether an item (possibly a nested list etc) contains a variable
    (any subtype of Tensor) """
    # in case of nested lists/tuples, recursively call the function on it
    if isinstance(item, slice):
        return False
    elif isinstance(item, list) or isinstance(item, tuple):
        return numpy.sum([isvar(value) for value in item])
    # otherwise cheack that it is a subtype of Tensor or a Tracer and not
    # a callable
    else:
        cond1 = isinstance(item, Tensor) or type(item) in [Constant, OpTuple]
        #        cond2 = isinstance(item, jax.interpreters.partial_eval.JaxprTracer)
        cond3 = callable(item)
        return cond1 and not cond3  # (cond1 or cond2) and cond3


_numpy_signature_re = re.compile(r"^([\w., ]+=)?\s*[\w\.]+\(.*\)$")


def update_numpydoc(docstr, fun, op):
    """Transforms the numpy docstring to remove references of
       parameters that are supported by the numpy version but not the JAX version"""

    # Some numpy functions have an extra tab at the beginning of each line,
    # If this function is one of those we remove this extra tab from all the lines
    if not hasattr(op, "__code__"):
        return docstr
    if docstr[:4] == "    ":
        lines = docstr.split("\n")
        for idx, line in enumerate(lines):
            lines[idx] = line.replace("    ", "", 1)
        docstr = "\n".join(lines)

    begin_idx = docstr.find("Parameters")
    begin_idx = docstr.find("--\n", begin_idx) + 2
    end_idx = docstr.find("Returns", begin_idx)

    parameters = docstr[begin_idx:end_idx]
    param_list = parameters.replace("\n    ", "@@").split("\n")
    for idx, p in enumerate(param_list):
        param = p[: p.find(" : ")].split(", ")[0]
        if param not in op.__code__.co_varnames:
            param_list[idx] = ""
    param_list = [param for param in param_list if param != ""]
    parameters = "\n".join(param_list).replace("@@", "\n    ")
    return docstr[: begin_idx + 1] + parameters + docstr[end_idx - 2 :]


def get_output_tree(
    jax_function, *args, **kwargs,
):

    # we need to remove the static arguments first
    # we first do it for the kwars
    static_kwargs = {}
    var_kwargs = {}
    for name, arg in list(kwargs.items()):
        if not isvar(arg):
            static_kwargs.update({name: arg})
        else:
            var_kwargs.update({name: arg})

    # we need to do the same for the args
    who_static = [int(not isvar(arg)) for arg in args]
    static_args = [arg for i, arg in zip(who_static, args) if i]
    var_args = [arg for i, arg in zip(who_static, args) if not i]

    # we need to define an abstract function that only takes as input the
    # non-static arguments, internally join them with the static ones
    # and return the output. This is because the jax shape inference
    # functions does not work with static arguments (such as the dimensions
    # of the transpose function)

    def abstract_func(*args, **kwargs):
        all_args = _args_formatting(args, static_args, who_static)
        return jax_function(*all_args, **kwargs, **static_kwargs)

    # now we evaluate the shape from the jax built-in function
    tree = jax.eval_shape(abstract_func, *var_args, **var_kwargs)
    return tree


def jax_wrap(func, insert_default_kwargs=True, doc_func=None, is_method=False):
    if doc_func is None:
        doc_func = func

    @wraps(doc_func)
    def op(*args, seed=None, **kwargs):
        # now we evaluate the shape from the jax built-in function

        # if there is a name we remove it for now to use the jax tracer
        if "name" in kwargs:
            op_name = kwargs["name"]
            del kwargs["name"]
        else:
            op_name = None

        args = list2tuple(args)
        for i in kwargs:
            kwargs[i] = list2tuple(kwargs[i])

        # first we check if we are in a random function to be careful
        # with the key
        from . import random

        # this is just to get shape and dtype so we do not bother
        # to use the correct seed yet
        if func in random._RANDOM_FUNCTIONS:
            temp_args = (jax.random.PRNGKey(0),) + args
        else:
            temp_args = args

        tree = get_output_tree(func, *temp_args, **kwargs)

        # now we determine if it is an Op or a Tuple object based on the
        # infered shape
        if type(tree) == list or type(tree) == tuple:
            shapes = [t.shape for t in tree]
            dtypes = [t.dtype for t in tree]
            return OpTuple(
                *args,
                _jax_function=func,
                _shapes=shapes,
                _dtypes=dtypes,
                name=op_name,
                **kwargs,
            )
        elif func in random._RANDOM_FUNCTIONS:
            shape, dtype = tree.shape, tree.dtype
            return RandomOp(
                *args,
                _jax_function=func,
                _shape=shape,
                _dtype=dtype,
                _seed=seed,
                name=op_name,
                **kwargs,
            )
        else:
            shape, dtype = tree.shape, tree.dtype
            return Op(
                *args,
                _jax_function=func,
                _shape=shape,
                _dtype=dtype,
                name=op_name,
                **kwargs,
            )

    if not hasattr(func, "__doc__") or func.__doc__ is None:
        return op

    if doc_func is not None:
        sections = func.__doc__.split("\n\n")

        signatures = []
        summary = None
        for i in range(len(sections)):
            if _numpy_signature_re.match(sections[i]):
                signatures.append(sections[i])
            else:
                summary = sections[i].strip()
                break
        body = "\n\n".join(signatures + sections[i + 1 :])
        body = update_numpydoc(body, func, op)
        desc = "ADDITION"
        docstr = (
            "{summary}\n\nLAX-backend implementation of :func:`{fun}`.\n"
            "{lax_description}Original docstring below.\n\n{body}".format(
                summary=summary, lax_description=desc, fun=func.__name__, body=body,
            )
        )

        op.__name__ = func.__name__
        op.__doc__ = docstr

    return op


def wrap_class(c, method_exceptions=None):
    class meta:
        def __new__(cls, *args, **kwargs):

            # the first part consists into reexpressing any possible symjax
            # input into a jax one to first evaluate the class creator and
            # derive from its the equivalent symjax computational graph that
            # would produce the same class attributes
            new_args = []
            new_kwargs = {}
            for i in range(len(args)):
                if isinstance(args[i], Tensor):
                    new_args.append(jnp.zeros(args[i].shape, dtype=args[i].dtype))
                else:
                    new_args.append(args[i])
            for i in kwargs:
                if isinstance(kwargs[i], Tensor):
                    new_kwargs[i] = jnp.zeros(kwargs[i].shape, dtype=kwargs[i].dtype)
                else:
                    new_kwargs[i] = kwargs[i]

            # now we check which attributes were added during the class
            # creation, those are the ones that will have to be obtained from
            # a symjax computational graph based on the class inputs
            attr_before = c.__dict__.keys()
            instance = c(*new_args, **new_kwargs)
            attr_after = instance.__dict__.keys()

            news = [i for i in attr_after if i not in attr_before]
            news = [
                n
                for n in news
                if isinstance(instance.__dict__[n], jax.interpreters.xla.DeviceArray)
            ]

            # this function maps the class inputs to the creator generated
            # class attributes
            def function(*args, **kwargs):
                return [instance.__dict__[n] for n in news]

            init_op = jax_wrap(function)

            # we now allow our normal class creation to proceed
            obj = super().__new__(cls)
            obj._init_op = init_op
            obj._news = news

            # we also have to wrap all the methods
            method_exceptions = cls._method_exceptions or []
            for att in dir(instance):
                if att[:2] == "__" or att in method_exceptions:
                    continue
                if callable(getattr(instance, att)):
                    setattr(obj, att, jax_wrap(getattr(instance, att)))
            return obj

        def __init__(self, *args, **kwargs):
            attrs = self._init_op(*args, **kwargs)
            for n, a in zip(self._news, attrs):
                self.__dict__[n] = a

    meta._method_exceptions = method_exceptions

    return meta


class Tensor:

    __array_priority__ = 1000

    def __init__(self, _shape, _dtype, name=None, **kwargs):

        self._shape = tuple(_shape)
        self._dtype = jax.dtypes.dtype(_dtype)

        if name is not None:
            assert "/" not in name
            self._name = name
        else:
            self._name = "unnamed"

        symjax.current_graph().add(self, **kwargs)

    @property
    def name(self):
        return self._name

    def _set_name(self, new_name):
        self._name = new_name

    def clone(self, givens):
        return symjax.current_graph().clone(self, givens)

    @property
    def shape(self):
        return self._shape

    @property
    def dtype(self):
        return self._dtype

    @property
    def ndim(self):
        return len(self.shape)


class Constant:
    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return "ConstantValue({})".format(type(self.value))

    def __str__(self):
        return self.__repr__()


class Op(Tensor):
    """an Op generates a Tensor object obtained from a function"""

    def __init__(
        self, *args, _jax_function, _shape=None, _dtype=None, name=None, **kwargs,
    ):
        self._fn = _jax_function.__name__
        # if the _shape and _dtype is not given (thus this class is
        # instanciated from anything else than the jax_wrap call)
        # then we get shape and dtype manually
        if _shape is None and _dtype is None:
            tree = get_output_tree(_jax_function, *args, **kwargs)
            _shape, _dtype = tree.shape, tree.dtype

        assert _shape is not None and _dtype is not None

        if name is None:
            name = _jax_function.__name__

        super().__init__(
            _shape,
            _dtype,
            name=name,
            args=args,
            kwargs=kwargs,
            jax_function=_jax_function,
        )

    def __repr__(self):

        name = "Op(name={}, fn={}, shape={}, dtype={}, scope={})"
        return name.format(self.name, self._fn, self.shape, self.dtype, self.scope)

    def __str__(self):

        return self.__repr__()


class OpTuple:
    def __init__(self, *args, _jax_function, _shapes, _dtypes, name=None, **kwargs):

        if name is None:
            name = _jax_function.__name__
        self._name = name
        symjax.current_graph().add(
            self, jax_function=_jax_function, args=args, kwargs=kwargs, **kwargs,
        )
        for i, (shape, dtype) in enumerate(zip(_shapes, _dtypes)):
            OpTupleItem(
                shape, dtype, index=i, parent=self, name=name + "[{}]".format(i),
            )

    def __repr__(self):
        successors = symjax.current_graph().successors(self)

        return "(" + ", ".join([str(a) for a in successors]) + ")"

    @property
    def name(self):
        return self._name

    def _set_name(self, name):
        self._name = name

    def __str__(self):

        return self.__repr__()

    def __iter__(self):
        """ Returns the Iterator object """
        return symjax.current_graph().successors(self)

    def __getitem__(self, item):
        assert type(item) == int
        for node in symjax.current_graph().successors(self):
            if symjax.current_graph()[self][node]["index"] == item:
                return node


class OpTupleItem(Tensor):
    def __init__(self, shape, dtype, index, parent, name):
        super().__init__(shape, dtype, name=name, parent=parent, index=index)

    def __repr__(self):

        name = "OpTupleItem(name={}, shape={}, dtype={}, scope={})"
        return name.format(self.name, self.shape, self.dtype, self.scope)

    def __str__(self):

        return self.__repr__()


class RandomOp(Op):
    """
    This class creates a :obj:`Tensor` object that given a function (see below)
    and its inputs can be used as a Node in the graph construction. This class
    is specialized to deal with random functions, if the function does not
    take a jax.PRNGKey as first argument, then it should not be used.

    Notes:
        This class is not meant to be used by the user. To create your own
        callable node, see :class:`RandomOp`.

    Args:
        msg (str): Human readable string describing the exception.
        code (:obj:`int`, optional): Error code.

    Attributes:
        msg (str): Human readable string describing the exception.
        code (int): Exception error code.

    """

    def __init__(self, *args, _jax_function, _shape, _dtype, _seed, name, **kwargs):
        self._shape = _shape
        self._dtype = _dtype
        self._fn = _jax_function.__name__
        super().__init__(
            *args,
            _jax_function=_jax_function,
            _shape=_shape,
            _dtype=_dtype,
            name=name,
            _seed=_seed,
            **kwargs,
        )

    def __repr__(self):
        name = "RandomTensor(name={}, fn={}, shape={}, dtype={}, scope={})"
        return name.format(self.name, self._fn, self.shape, self.dtype, self.scope)


class Variable(Tensor):

    """variable that is a standalone persistent tensor. this tensor
    can be updated and differentiated.

    Parameters:
    -----------

        value_or_fn: array or initializer
            the value given as a numpy array or an initializer which
            takes as input the shape and can be type casted afterward
            via numpy.cast

        shape: tuple (optional)
            the shape of the variable, used only if the value_or_fn is an
            initializer

        dtype: dtype (optional)
            the dtype of the variable, used only if the value_or_fn is an
            initializer

        name: str (optional)
            the name of the variable, there is no test of name duplication

        trainable: bool
            whether the variable is trainable or not. It is set as an
            attribute and can be accessed.
    """

    def __init__(
        self, initializer, name="unnamed_variable", trainable=True, dtype=None
    ):

        if trainable and dtype == "bool":
            raise RuntimeError("error impossible learning with dtype bool")
        self.trainable = trainable
        self.initializer = initializer
        self._dtype = dtype
        super().__init__(self.value.shape, str(self.value.dtype), name=name)

    def reset(self):
        """reset the value of the variable based on the initial one, whether
        it was an array or initializer. If it was a random initializer,
        nothing guarantees that the reset will give back the original value
        as opposed to the array case
        """
        if isinstance(self.initializer, Tensor):
            self._value = symjax.current_graph().get(self.initializer)
        else:
            self._value = numpy.array(self.initializer)

        if self._dtype is not None:
            self._value = self._value.astype(self._dtype)

    @property
    def value(self):
        """ utility function that takes the input and return
            the actual value. It deals with cases where the input
            was a function or not etc
        """
        if not hasattr(self, "_value"):
            self.reset()
        return self._value

    def update(self, update_value):
        """assign a new value for the variable"""
        new_value = symjax.current_graph().get(update_value)

        if self.shape != jax.numpy.shape(new_value):
            warnings.warn(
                "Variable and update {} {}".format(self, new_value)
                + "are not the same shape... attempting to reshape"
            )
            new_value = jax.numpy.reshape(new_value, self.shape)

        if hasattr(new_value, "dtype"):
            ntype = new_value.dtype
        else:
            ntype = type(new_value)
        if self.dtype != ntype:
            warnings.warn(
                "Variable and update {}and {}".format(self, new_value)
                + "are not the same dtype... attempting to cast"
            )

            new_value = jax.numpy.asarray(new_value).astype(self.dtype)

        self._value = new_value

    def __repr__(self):
        name = "Variable(name={}, shape={}, dtype={}, trainable={}, scope={})"
        return name.format(
            self.name, self.shape, self.dtype, self.trainable, self.scope
        )


class Placeholder(Tensor):
    """placeholder is an input to the computational graph that takes outside
    values. That is, it is an input gate to feed data into a computational
    graph as opposed to for example variables which are living in memory and
    are not fed externally.

    Parameters:
    -----------

        shape: tuple
            the shape of the placeholder

        dtype: dtype
            the dtype of the placeholder

        name: str (optional)
            the name of the variable, there is no test of name duplication
    """

    def __init__(self, shape, dtype, name="unnamed"):
        super().__init__(shape, dtype, name=name)

    def __repr__(self):
        name = "Placeholder(name={}, shape={}, dtype={}, scope={})"
        return name.format(self.name, self.shape, self.dtype, self.scope)


def placeholder_like(item, name=""):
    if item is None:
        return None
    elif type(item) == list or type(item) == tuple:
        return type(item)([placeholder_like(i) for i in item])
    else:
        return Placeholder(item.shape, item.dtype, name=name)


def match(l1, l2, output):
    if output is None:
        output = dict()
    if type(l1) == list or type(l1) == tuple:
        for a, b in zip(l1, l2):
            match(a, b, output)
    else:
        output.update({l1: l2})


def symjax_to_jax_fn(func):
    def newfn(*args, fn=func):
        pholders = placeholder_like(args)
        symjax_outputs = fn(*pholders)
        feed_dict = {}
        match(pholders, args, feed_dict)
        if None in feed_dict:
            del feed_dict[None]
        outputs = symjax.current_graph().get(symjax_outputs, feed_dict)
        return outputs

    return newfn


def clone(tensor, givens):
    return tensor.clone(givens)


def get(tensor, tracker=None):
    return symjax.current_graph().get(tensor, tracker)



import os
import numpy as np
import time
from . import Dataset


from sklearn.datasets import make_moons, make_circles
from sklearn.model_selection import train_test_split


def load_path(N=1000, std=0.03, option=0, seed=None):
    X = np.linspace(0, 1, N)
    if option == 0:
        # simple cosine
        X = np.stack([X * 9, np.cos(X * 9)], 1)
    elif option == 1:
        # circle
        X -= 0.5
        X *= 2
        Y = np.concatenate([np.sqrt(1 - X[::2] ** 2), -np.sqrt(1 - X[::2] ** 2)], 0)
        X = np.concatenate([X[::2], X[::2]], 0)
        X = np.stack([X, Y], 1)
    elif option == 2:
        # swiss roll
        t = np.linspace(0, 2, N)
        X = np.stack([t * np.cos(3.14 * t), t * np.sin(3.14 * t)], 1)
    elif option == 3:
        # square
        left_side = np.stack([np.ones(N // 4) * X[0], X[::4]], 1)
        right_side = np.stack([np.ones(N // 4) * X[-1], X[::4]], 1)
        upper_side = np.stack([X[::4], np.ones(N // 4) * X[0]], 1)
        lower_side = np.stack([X[::4], np.ones(N // 4) * X[-1]], 1)
        X = np.concatenate([left_side, upper_side, right_side, lower_side[::-1]], 0)
    elif option == 4:
        # simple cosine
        X = np.stack([X * 9, np.random.rand(*X.shape), np.cos(X * 9)], 1)
    elif option == 5:
        #        X = np.random.randn(N, 3)
        #        X /= np.linalg.norm(X, 2, axis=1, keepdims=True)
        #        X[:,0] = np.abs(X[:, 0])
        import sklearn

        X, _ = sklearn.datasets.make_swiss_roll(N, 0)
    X += np.random.RandomState(seed).randn(N, X.shape[1]) * std
    X -= X.mean(0, keepdims=True)
    X /= X.max(0, keepdims=True)

    X = X.astype("float32")

    dataset = Dataset()
    dataset["inputs/train_set"] = X
    return dataset


def load_mini(N=1000):
    X, y = make_moons(N, noise=0.035, random_state=20)
    x_, y_ = make_circles(N, noise=0.02, random_state=20)
    x_[:, 1] += 2.0
    y_ += 2
    X = np.concatenate([X, x_], axis=0)
    y = np.concatenate([y, y_])
    X -= X.mean(0, keepdims=True)
    X /= X.max(0, keepdims=True)

    X = X.astype("float32")
    y = y.astype("int32")

    dict_init = [
        ("datum_shape", (2,)),
        ("n_classes", 4),
        ("name", "mini"),
        ("classes", [str(u) for u in range(4)]),
    ]

    dataset = Dataset(**dict(dict_init))
    dataset["inputs/train_set"] = X
    dataset["outputs/train_set"] = y

    return dataset


def load_chirp2D(N, seed=0):
    X1, X2 = np.meshgrid(np.linspace(0, 4, N), np.linspace(0, 4, N))
    np.random.seed(seed)
    M = np.array([[1.4, -0.4], [-0.4, 0.6]])
    X = np.stack([X1.flatten(), X2.flatten()], 1)
    y = np.sin((X * np.dot(X, M)).sum(1))
    y -= y.mean()
    y /= y.max()

    X = X.astype("float32")
    y = y.astype("float32")

    dict_init = [("datum_shape", (2,)), ("name", "chirp2D")]

    dataset = Dataset(**dict(dict_init))

    images = {"train_set": X}

    labels = {"train_set": y}

    dataset.add_variable({"input": images, "output": labels})

    return dataset



import sys

import jax.random as jnp

from .base import jax_wrap

_seed = 0

_RANDOM_FUNCTIONS = [
    jnp.bernoulli,
    jnp.beta,
    jnp.categorical,
    jnp.cauchy,
    jnp.dirichlet,
    jnp.exponential,
    jnp.gamma,
    jnp.gumbel,
    jnp.laplace,
    jnp.logistic,
    jnp.multivariate_normal,
    jnp.normal,
    jnp.pareto,
    jnp.permutation,
    jnp.poisson,
    jnp.randint,
    jnp.shuffle,
    jnp.t,
    jnp.threefry_2x32,
    jnp.truncated_normal,
    jnp.uniform,
]

module = sys.modules[__name__]

for name in _RANDOM_FUNCTIONS:
    module.__dict__.update({name.__name__: jax_wrap(name)})

randn = module.__dict__["normal"]
rand = module.__dict__["uniform"]



import os
import pickle, gzip
import urllib.request
import numpy as np
import time
import tarfile
from tqdm import tqdm
from scipy.io.wavfile import read as wav_read


class sonycust:
    """multilabel urban sound classification

    Reference at https://zenodo.org/record/3233082 

    Description

    SONYC Urban Sound Tagging (SONYC-UST) is a dataset for the development and
    evaluation of machine listening systems for realistic urban noise
    monitoring. The audio was recorded from the SONYC acoustic sensor network.
    Volunteers on the Zooniverse citizen science platform tagged the presence
    of 23 classes that were chosen in consultation with the New York City
    Department of Environmental Protection. These 23 fine-grained classes can be
    grouped into 8 coarse-grained classes. The recordings are split into three
    subsets: training, validation, and test. These sets are disjoint with
    respect to the sensor from which each recording came. For increased
    reliability, three volunteers annotated each recording, and members of the
    SONYC team subsequently created a set of ground-truth tags for the
    validation set using a two-stage annotation procedure in which two
    annotators independently tagged and then collectively resolved any
    disagreements. For more details on the motivation and creation of this
    dataset see the DCASE 2019 Urban Sound Tagging Task website.

    Audio data
    
    The provided audio has been acquired using the SONYC acoustic sensor network 
    for urban noise pollution monitoring. Over 50 different sensors have been
    deployed in New York City, and these sensors have collectively gathered the
    equivalent of 37 years of audio data, of which we provide a small subset.
    The data was sampled by selecting the nearest neighbors on VGGish features
    of recordings known to have classes of interest. All recordings are 10
    seconds and were recorded with identical microphones at identical gain
    settings. To maintain privacy, the recordings in this release have been
    distributed in time and location, and the time and location of the
    recordings are not included in the metadata.

    Labels

    there are fine and coarse labels
    engine
    1: small-sounding-engine
    2: medium-sounding-engine
    3: large-sounding-engine
    X: engine-of-uncertain-size
    machinery-impact
    1: rock-drill
    2: jackhammer
    3: hoe-ram
    4: pile-driver
    X: other-unknown-impact-machinery
    non-machinery-impact
    1: non-machinery-impact
    powered-saw
    1: chainsaw
    2: small-medium-rotating-saw
    3: large-rotating-saw
    X: other-unknown-powered-saw
    alert-signal
    1: car-horn
    2: car-alarm
    3: siren
    4: reverse-beeper
    X: other-unknown-alert-signal
    music
    1: stationary-music
    2: mobile-music
    3: ice-cream-truck
    X: music-from-uncertain-source
    human-voice
    1: person-or-small-group-talking
    2: person-or-small-group-shouting
    3: large-crowd
    4: amplified-speech
    X: other-unknown-human-voice
    dog
    1: dog-barking-whining

    """

    coarse_labels = [
        "engine",
        "machinery-impact",
        "non-machinery-impact" "powered-saw",
        "alert-signal",
        "music",
        "human-voice",
        "dog",
    ]

    fine_labels = [
        "small-sounding-engine",
        "medium-sounding-engine",
        "large-sounding-engine",
        "engine-of-uncertain-size",
        "rock-drill",
        "jackhammer",
        "hoe-ram",
        "pile-driver",
        "other-unknown-impact-machinery",
        "non-machinery-impact",
        "chainsaw",
        "small-medium-rotating-saw",
        "large-rotating-saw",
        "other-unknown-powered-saw",
        "car-horn",
        "car-alarm",
        "siren",
        "reverse-beeper",
        "other-unknown-alert-signal",
        "stationary-music",
        "mobile-music",
        "ice-cream-truck",
        "music-from-uncertain-source",
        "person-or-small-group-talking",
        "person-or-small-group-shouting",
        "large-crowd",
        "amplified-speech",
        "other-unknown-human-voice",
        "dog-barking-whining",
    ]

    def download(path):

        # Check if directory exists
        if not os.path.isdir(path + "ust"):
            print("Creating mnist Directory")
            os.mkdir(path + "ust")

        url = "https://zenodo.org/record/3233082/files/audio-dev.tar.gz?download=1"
        # Check if file exists
        if not os.path.exists(path + "ust/audio-dev.tar.gz"):
            td = time.time()
            urllib.request.urlretrieve(url, path + "ust/audio-dev.tar.gz")

        url = "https://zenodo.org/record/3233082/files/annotations-dev.csv?download=1"
        # Check if file exists
        if not os.path.exists(path + "ust/annotations-dev.csv"):
            td = time.time()
            urllib.request.urlretrieve(url, path + "ust/annotations-dev.csv")

    def load(path=None, classes=range(10)):

        if path is None:
            path = os.environ["DATASET_PATH"]
        sonycust.download(path)

        t0 = time.time()

        # Loading the file
        files = tarfile.open(path + "ust/audio-dev.tar.gz", "r:gz")
        annotations = np.loadtxt(
            path + "ust/annotations-dev.csv", delimiter=",", skiprows=1, dtype="str"
        )

        # get name
        filenames = list(annotations[:, 2])
        for i in range(len(filenames)):
            filenames[i] = annotations[i, 0] + "/" + str(filenames[i])

        # get fine labels and limts for coarse classes
        fine_labels = annotations[:, 4:33].astype("float32").astype("int32")
        class_limits = [0, 4, 9, 10, 14, 19, 23, 28, 29]
        n_classes = len(class_limits) - 1
        n_samples = len(annotations)
        llabels = np.zeros((n_samples, n_classes), dtype="int")
        for k in range(n_classes):
            block = fine_labels[:, class_limits[k] : class_limits[k + 1]]
            llabels[:, k] = block.max(1)

        wavs = np.zeros((2794, 441000), dtype="float32")
        coarse = np.zeros((2794, 8), dtype="int32")
        fine = np.zeros((2794, 29), dtype="int32")
        filenames = files.getnames()
        cpt = 0
        for name in tqdm(filenames, ascii=True):
            if ".wav" not in name:
                continue
            wav = wav_read(files.extractfile(name))[1].astype("float32")
            wavs[cpt] = wav_read(files.extractfile(name))[1].astype("float32")
            coarse[cpt] = llabels[filenames.index(name)]
            fine[cpt] = fine_labels[filenames.index(name)]
            cpt += 1
        return wavs, fine, coarse



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__all__ = ["random", "signal", "linagl", "fft", "interpolation"]


from .base import *


from .control_flow import map, scan, while_loop

from .ops_numpy import *
from .ops_special import *

from . import random
from . import signal
from . import linalg
from . import fft
from . import interpolation



import os
import zipfile
import io
from tqdm import tqdm
import urllib.request
import numpy as np
import time
from scipy.io.wavfile import read as wav_read


class irmas:
    """music instrument classification

    ref https://zenodo.org/record/1290750#.WzCwSRyxXMU

    This dataset includes musical audio excerpts with annotations of the
    predominant instrument(s) present. It was used for the evaluation in the
    following article:

    Bosch, J. J., Janer, J., Fuhrmann, F., & Herrera, P. “A Comparison of Sound
    Segregation Techniques for Predominant Instrument Recognition in Musical
    Audio Signals”, in Proc. ISMIR (pp. 559-564), 2012

    Please Acknowledge IRMAS in Academic Research

    IRMAS is intended to be used for training and testing methods for the
    automatic recognition of predominant instruments in musical audio. The
    instruments considered are: cello, clarinet, flute, acoustic guitar,
    electric guitar, organ, piano, saxophone, trumpet, violin, and human singing
    voice. This dataset is derived from the one compiled by Ferdinand Fuhrmann
    in his PhD thesis, with the difference that we provide audio data in stereo
    format, the annotations in the testing dataset are limited to specific
    pitched instruments, and there is a different amount and lenght of excerpts.
    """

    def download(path):
        # Check if directory exists
        if not os.path.isdir(path + "irmas"):
            print("\tCreating irmas Directory")
            os.mkdir(path + "irmas")

        base = "https://zenodo.org/record/1290750/files/"
        # Check if file exists
        if not os.path.exists(path + "irmas/IRMAS-TrainingData.zip"):
            print("\tDownloading training data")
            url = base + "IRMAS-TrainingData.zip?download=1"
            urllib.request.urlretrieve(url, path + "irmas/IRMAS-TrainingData.zip")

        # Check if file exists
        if not os.path.exists(path + "irmas/IRMAS-TestingData-Part1.zip"):
            print("\tDownloading test data 1/3")
            url = base + "IRMAS-TestingData-Part1.zip?download=1"
            target = "irmas/IRMAS-TestingData-Part1.zip"
            urllib.request.urlretrieve(url, path + target)

        # Check if file exists
        if not os.path.exists(path + "irmas/IRMAS-TestingData-Part2.zip"):
            print("\tDownloading test data 2/3")
            url = base + "IRMAS-TestingData-Part2.zip?download=1"
            target = "irmas/IRMAS-TestingData-Part2.zip"
            urllib.request.urlretrieve(url, path + target)

        # Check if file exists
        if not os.path.exists(path + "irmas/IRMAS-TestingData-Part3.zip"):
            print("\tDownloading test data3/3")
            url = base + "IRMAS-TestingData-Part3.zip?download=1"
            target = "irmas/IRMAS-TestingData-Part3.zip"
            urllib.request.urlretrieve(url, path + target)

    def load(path=None):

        if path is None:
            path = os.environ["DATASET_PATH"]
        irmas.download(path)

        t0 = time.time()

        train_wavs = list()
        train_labels = list()
        test_wavs = list()
        test_labels = list()

        # loading the training set
        f = zipfile.ZipFile(path + "irmas/IRMAS-TrainingData.zip")
        namelist = f.namelist()
        for filename in tqdm(namelist, ascii=True):
            if ".wav" not in filename:
                continue
            wavfile = f.read(filename)
            byt = io.BytesIO(wavfile)
            train_wavs.append(wav_read(byt)[1].astype("float32"))
            train_labels.append(filename.split("/")[-2])

        base = "irmas/IRMAS-TestingData-Part{}.zip"
        for part in ["1", "2", "3"]:
            f = zipfile.ZipFile(path + base.format(part))
            namelist = f.namelist()
            for filename in tqdm(
                namelist, ascii=True, desc="Test data {}/3".format(part)
            ):
                if ".wav" not in filename:
                    continue

                byt = io.BytesIO(f.read(filename))
                test_wavs.append(wav_read(byt)[1].astype("float32"))

                byt = io.BytesIO(f.read(filename.replace(".wav", ".txt")))
                test_labels.append(np.loadtxt(byt, dtype="str"))

        unique_cat = np.unique(categories)
        Id = np.eye(len(unique_cat))
        train_labels = np.array(
            [Id[unique_cat.index(cat)] for cat in train_labels]
        ).astype("int32")
        test_labels = np.array(
            [Id[unique_cat.index(cat)] for cat in test_labels]
        ).astype("int32")

        train_wavs = np.array(train_wavs)
        test_wavs = np.array(test_wavs)

        data = {
            "train_set/wavs": train_wavs,
            "train_set/labels": train_labels,
            "test_wavs": test_wavs,
            "test_labels": test_labels,
            "INFOS": irmas.__doc__,
        }

        print("Dataset IRMAS loaded in {0:.2f}s.".format(time.time() - t0))
        return data



# Based on a script by Seiya Tokui. With the following copyright
# Copyright (c) 2014 Seiya Tokui
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# Given the wnid of a synset, the wnid of hyponym synsets can be obtained at
# http://www.image-net.org/api/text/wordnet.structure.hyponym?wnid=[wnid]
#
# To obtain the full hyponym (the synset of the whole subtree starting
# from wnid), you can request
# http://www.image-net.org/api/text/wordnet.structure.hyponym?wnid=[wnid]&full=1
#
# to get the word of s synset
# http://www.image-net.org/api/text/wordnet.synset.getwords?wnid=[wnid]
#
# Given the wnid of a synset, the URLs of its images can be obtained at
# http://www.image-net.org/api/text/imagenet.synset.geturls?wnid=[wnid]
#
# mappingfrom all synset to words
# http://image-net.org/archive/words.txt
#


import argparse
import urllib.request, urllib.error, urllib.parse
import time
import os
import math
import threading
import sys
import imghdr
import http.client
from ssl import CertificateError


class DownloadError(Exception):
    """Base class for exceptions in this module."""

    def __init__(self, message=""):
        self.message = message


def download(n_images, min_size, n_threads, wnids_list, out_dir):
    wnid_thread_lists = list()
    wnid_list_len = len(wnids_list)
    wnid_thread_sizes = int(math.ceil(float(wnid_list_len) / n_threads))
    for i in range(n_threads):
        wnid_thread_lists.append(
            wnids_list[i * wnid_thread_sizes : (i + 1) * wnid_thread_sizes]
        )

    # Define the threads
    def downloader(wnid_list):
        for wnid in wnid_list:
            dir_name = wnid
            print("Downloading " + dir_name)
            dir_path = os.path.join(out_dir, dir_name)
            if os.path.isdir(dir_path):
                print("skipping: already have " + dir_name)
            else:
                image_url_list = get_image_urls(wnid)
                download_images(dir_path, image_url_list, n_images, min_size)

    # initialize the threads
    print(wnid_thread_lists[0])
    download_threads = [
        threading.Thread(target=downloader, args=([wnid_thread_lists[i]]))
        for i in range(n_threads)
    ]

    for t in download_threads:
        t.start()

    is_alive = True
    while is_alive:
        is_alive = False
        for t in download_threads:
            is_alive = is_alive or t.isAlive()
        time.sleep(0.1)

    for t in download_threads:
        t.join()
    print("finished")


def mkdir(path):
    if not os.path.isdir(path):
        os.makedirs(path)


def get_url_request_list_function(request_url):
    def get_url_request_list(wnid, timeout=5, retry=3):
        url = request_url + wnid
        f = urllib.request.urlopen(url)
        response = f.read().decode()
        f.close()
        print("response: " + response)
        list = str.split(response)
        return list

    return get_url_request_list


get_image_urls = get_url_request_list_function(
    "http://www.image-net.org/api/text/imagenet.synset.geturls?wnid="
)

get_subtree_wnid = get_url_request_list_function(
    "http://www.image-net.org/api/text/wordnet.structure.hyponym?wnid="
)

get_full_subtree_wnid = get_url_request_list_function(
    "http://www.image-net.org/api/text/wordnet.structure.hyponym?full=1&wnid="
)


def get_words_wnid(wnid):
    url = "http://www.image-net.org/api/text/wordnet.synset.getwords?wnid=" + wnid
    f = urllib.request.urlopen(url)
    content = f.read().decode()
    f.close()
    return content


def download_images(dir_path, image_url_list, n_images, min_size):
    mkdir(dir_path)
    image_count = 0
    for url in image_url_list:
        if image_count == n_images:
            break
        try:
            f = urllib.request.urlopen(url)
            image = f.read()
            f.close()
            extension = imghdr.what("", image)  # check if valid image
            if extension == "jpeg":
                extension = "jpg"
            if sys.getsizeof(image) > min_size:
                image_name = "image_" + str(image_count) + "." + extension
                image_path = os.path.join(dir_path, image_name)
                image_file = open(image_path, "wb")
                image_file.write(image)
                image_file.close()
                image_count += 1
        except:
            print("skipping ", url)


def main(wnid, out_dir, n_threads, n_images, fullsubtree, noroot, nosubtree, min_size):

    wnids_list = []

    # First get the list of wnids
    if not noroot:
        wnids_list.append(wnid)
    if not nosubtree:
        if fullsubtree:
            subtree = get_full_subtree_wnid(wnid)
        else:
            subtree = get_subtree_wnid(wnid, timeout, retry)
        for i in range(1, len(subtree)):
            subtree[i] = subtree[i][1:]  # removes dash
        wnids_list.extend(subtree)

    # create root directory
    mkdir(out_dir)
    download(n_images, min_size, n_threads, wnids_list, out_dir)


if __name__ == "__main__":
    p = argparse.ArgumentParser()
    p.add_argument("wnid", help="Imagenet wnid, example n03489162")
    p.add_argument("outdir", help="Output directory")
    p.add_argument(
        "--jobs",
        "-j",
        type=int,
        default=1,
        help="Number of parallel threads to download",
    )
    p.add_argument(
        "--images",
        "-i",
        type=int,
        default=20,
        metavar="N_IMAGES",
        help="Number of images per category to download",
    )
    p.add_argument(
        "--fullsubtree", "-F", action="store_true", help="Downloads the full subtree"
    )
    p.add_argument(
        "--noroot", "-R", action="store_true", help="Do not Downloads the root"
    )
    p.add_argument(
        "--nosubtree", "-S", action="store_true", help="Do not Downloads the subtree"
    )

    p.add_argument(
        "--humanreadable",
        "-H",
        action="store_true",
        help="Makes the folders human readable",
    )

    p.add_argument(
        "--minsize",
        "-m",
        type=float,
        default=7000,
        help="Min size of the images in bytes",
    )

    args = p.parse_args()
    main(
        wnid=args.wnid,
        out_dir=args.outdir,
        n_threads=args.jobs,
        n_images=args.images,
        fullsubtree=args.fullsubtree,
        noroot=args.noroot,
        nosubtree=args.nosubtree,
        min_size=args.minsize,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Randall Balestriero"

import os
import io
import urllib.request
import numpy as np
import time
from tqdm import tqdm
import zipfile
from scipy.io.wavfile import read as wav_read


__INFO__ = """An Annotated Acoustic Dataset of 7 Picidae Species

    The proposed dataset contains 1669 labeled audio files from
    the following Picidae species doing 3 types of birdsongs:
    call, drumming and song. Audio data are organized in thirteen
    folders: twelve containing the labeled audios of each
    birdsong type from every Picidae specie listed in what
    follows and one folder containing the background sound samples

    1- DendrocoposLeucotos - call.

    2- DendrocoposLeucotos - drumming.

    3- DendrocoposMajor - call.

    4- DendrocoposMajor - drumming.

    5- DendrocoposMedius - call.

    6- DendrocoposMedius - song.

    7- DendrocoposMinor - call.

    8- DendrocoposMinor - drumming.

    9- DryocopusMartius - call.

    10- DryocopusMartius - drumming.

    11- JynxTorquilla - song.

    12- PicusViridis - song.

    13- Silence (or background noise).

   """


def download(path):
    """
    Download the dataset and store the result into the given
    path

    Parameters
    ----------

        path: str
            the path where the downloaded files will be stored. If the
            directory does not exist, it is created.
    """

    # Check if directory exists
    if not os.path.isdir(path + "picidae"):
        print("Creating picidae Directory")
        os.mkdir(path + "picidae")
    url = "https://zenodo.org/record/574438/files/PicidaeDataset.zip?download=1"
    if not os.path.exists(path + "picidae/PicidaeDataset.zip"):
        urllib.request.urlretrieve(url, path + "picidae/PicidaeDataset.zip")


def load(path=None):
    """
    Parameters
    ----------
        path: str (optional)
            default ($DATASET_PATH), the path to look for the data and
            where the data will be downloaded if not present

    Returns
    -------

        wavs: array
            the waveforms in the time amplitude domain

        labels: array
            binary values representing the presence or not of an avian

        flag: array
            the Xeno-Canto ID

    """

    if path is None:
        path = os.environ["DATASET_PATH"]

    download(path)

    t0 = time.time()

    archive = zipfile.ZipFile(path + "picidae/PicidaeDataset.zip")
    wavs = list()
    labels = list()
    XC = list()
    for item in tqdm(archive.namelist(), ascii=True):
        if item[-4:] == ".wav" and "._" not in item:
            wavfile = archive.read(item)
            byt = io.BytesIO(wavfile)
            wavs.append(wav_read(byt)[1].astype("float32"))
            labels.append(item.split("/")[1])
            XC.append(item.split("/")[2].split("-")[0])

    labels = np.array(labels)
    unique = np.unique(labels)
    y = np.zeros(len(labels), dtype="int32")
    for k, name in enumerate(np.sort(unique)):
        y[labels == name] = k

    data = {
        "wavs": wavs,
        "labels": y,
        "names": labels,
        "XC_identifiers": XC,
        "INFOS": __INFO__,
    }

    print("Dataset picidae loaded in {0:.2f}s.".format(time.time() - t0))

    return data



from tqdm import tqdm
import matplotlib.image as mpimg
import zipfile
import numpy as np
import os
import requests
import time
import urllib.request
from urllib.request import urlretrieve


def download_file_from_google_drive(file_id, dest, md5=None):
    """Download a Google Drive file from  and place it in root.
    Args:
        file_id (str): id of file to be downloaded
        dist (str): name of file to be saved
        md5 (str, optional): MD5 checksum of the download. If None, do not check
        :param dest:
    """
    # Based on https://stackoverflow.com/questions/38511444/python-download-files-from-google-drive-using-url
    import requests

    url = "https://docs.google.com/uc?export=download"

    if os.path.isfile(dest) and check_integrity(dest, md5):
        print("Using downloaded and verified file: " + fpath)
    else:
        session = requests.Session()

        response = session.get(url, params={"id": file_id}, stream=True)
        token = _get_confirm_token(response)

        if token:
            params = {"id": file_id, "confirm": token}
            response = session.get(url, params=params, stream=True)

        _save_response_content(response, dest)


def _get_confirm_token(response):
    for key, value in response.cookies.items():
        if key.startswith("download_warning"):
            return value
    return None


def _save_response_content(response, destination, chunk_size=32768):
    with open(destination, "wb") as f:
        pbar = tqdm(total=None)
        progress = 0
        for chunk in response.iter_content(chunk_size):
            if chunk:  # filter out keep-alive new chunks
                f.write(chunk)
                progress += len(chunk)
                pbar.update(progress - pbar.n)
        pbar.close()


class DLProgress(tqdm):
    """
    Handle Progress Bar while Downloading
    """

    last_block = 0

    def hook(self, block_num=1, block_size=1, total_size=None):
        """
        A hook function that will be called once on establishment of the network connection and
        once after each block read thereafter.
        :param block_num: A count of blocks transferred so far
        :param block_size: Block size in bytes
        :param total_size: The total size of the file. This may be -1 on older FTP servers which do not return
                            a file size in response to a retrieval request.
        """
        self.total = total_size
        self.update((block_num - self.last_block) * block_size)
        self.last_block = block_num


def download_extract(path):
    """
    Download and extract database
    :param path:
    """
    url = "https://s3-us-west-1.amazonaws.com/udacity-dlnfd/datasets/celeba.zip"
    hash_code = "00d2c5bc6d35e252742224ab0c1e8fcb"
    save_path = os.path.join(path, "celeba.zip")

    if not os.path.exists(path):
        os.makedirs(path)

    if not os.path.exists(save_path):
        with DLProgress(
            unit="B", unit_scale=True, miniters=1, desc="Downloading celebA"
        ) as pbar:
            urlretrieve(url, save_path, pbar.hook)


#    assert hashlib.md5(open(save_path, 'rb').read()).hexdigest() == hash_code, \
#        '{} file is corrupted.  Remove the file and try again.'.format(save_path)


def read_image_from_archive(arhive, image_name):
    imgfile = archive.open(image_name, "r")
    img = mpimg.imread(imgfile)
    return img


class celebA:
    """Celebrity faces.
    """

    file_list = [
        # File ID                         MD5 Hash                            Filename
        (
            "0B7EVK8r0v71pZjFTYXZWM3FlRnM",
            "00d2c5bc6d35e252742224ab0c1e8fcb",
            "img_align_celeba.zip",
        ),
        # ("0B7EVK8r0v71pbWNEUjJKdDQ3dGc", "b6cd7e93bc7a96c2dc33f819aa3ac651", "img_align_celeba_png.7z"),
        # ("0B7EVK8r0v71peklHb0pGdDl6R28", "b6cd7e93bc7a96c2dc33f819aa3ac651", "img_celeba.7z"),
        (
            "0B7EVK8r0v71pblRyaVFSWGxPY0U",
            "75e246fa4810816ffd6ee81facbd244c",
            "list_attr_celeba.txt",
        ),
        (
            "1_ee_0u7vcNLOfNLegJRHmolfH5ICW-XS",
            "32bd1bd63d3c78cd57e08160ec5ed1e2",
            "identity_CelebA.txt",
        ),
        (
            "0B7EVK8r0v71pbThiMVRxWXZ4dU0",
            "00566efa6fedff7a56946cd1c10f1c16",
            "list_bbox_celeba.txt",
        ),
        (
            "0B7EVK8r0v71pd0FJY3Blby1HUTQ",
            "cc24ecafdb5b50baae59b03474781f8c",
            "list_landmarks_align_celeba.txt",
        ),
        # ("0B7EVK8r0v71pTzJIdlJWdHczRlU", "063ee6ddb681f96bc9ca28c6febb9d1a", "list_landmarks_celeba.txt"),
        (
            "0B7EVK8r0v71pY0NSMzRuSXJEVkk",
            "d32c9cbf5e040fd4025c592c306e6668",
            "list_eval_partition.txt",
        ),
    ]

    @staticmethod
    def download(path):
        """
        Download the face dataset and store the result into the given
        path

        Parameters
        ----------

            path: str
                the path where the downloaded files will be stored. If the
                directory does not exist, it is created.
        """

        if not os.path.exists(path):
            os.makedirs(path)

        for file in celebA.file_list:
            save_path = os.path.join(path, file[-1])
            if not os.path.exists(save_path):
                download_file_from_google_drive(file[0], save_path)

    @staticmethod
    def load(path=None):
        """
        Parameters
        ----------
            path: str (optional)
                default ($DATASET_PATH), the path to look for the data and
                where the data will be downloaded if not present

        Returns
        -------

            train_images: array

            train_labels: array

            valid_images: array

            valid_labels: array

            test_images: array

            test_labels: array

        """

        if path is None:
            path = os.environ["DATASET_PATH"]

        celebA.download(path + "celebA/")

        t0 = time.time()

        archive = zipfile.ZipFile(path + "celebA/img_align_celeba.zip", "r")
        names = archive.namelist()
        print([name for name in names if name[-1] == "t"])
        names = [name for name in names if name[-3:] == "jpg"]
        names = names[:3000]
        data = np.empty((len(names), 218, 178, 3))
        attributes = np.loadtxt(path + "list_attr_celeba.txt")
        for i, name in enumerate(names):
            data[i] = read_image_from_archive(archive, name)

        print("Dataset celebA loaded in {0:.2f}s.".format(time.time() - t0))

        return data, attributes



import os
import tarfile
import time
import urllib.request
import matplotlib.image
from multiprocessing import Pool
from functools import partial

import numpy as np
from tqdm import tqdm


__DESCRIPTION__ = """Imagenette is a subset of 10 easily classified classes 
from Imagenet (tench, English springer, cassette player, chain saw, church, 
French horn, garbage truck, gas pump, golf ball, parachute).

'Imagenette' is pronounced just like 'Imagenet', except with a corny 
inauthentic French accent. If you've seen Peter Sellars in The Pink Panther,
 then think something like that. It's important to ham up the accent as much as
  possible, otherwise people might not be sure whether you're refering to 
  "Imagenette" or "Imagenet". (Note to native French speakers: to avoid
   confusion, be sure to use a corny inauthentic American accent when saying 
   "Imagenet". Think something like the philosophy restaurant skit from Monty
    Python's The Meaning of Life.)
"""


imagenette_map = {
    "n01440764": "tench",
    "n02102040": "springer",
    "n02979186": "casette_player",
    "n03000684": "chain_saw",
    "n03028079": "church",
    "n03394916": "French_horn",
    "n03417042": "garbage_truck",
    "n03425413": "gas_pump",
    "n03445777": "golf_ball",
    "n03888257": "parachute",
}


def download(path):
    """
    Download the MNIST dataset and store the result into the given
    path

    Parameters
    ----------

        path: str
            the path where the downloaded files will be stored. If the
            directory does not exist, it is created.
    """

    # Check if directory exists
    if not os.path.isdir(path + "imagenette"):
        print("\tCreating imagenette Directory")
        os.mkdir(path + "imagenette")

    # Check if file exists
    if not os.path.exists(path + "imagenette/imagenette2.tgz"):
        print("\tDownloading imagenette")
        url = "https://s3.amazonaws.com/fast-ai-imageclas/imagenette2.tgz"
        urllib.request.urlretrieve(url, path + "imagenette/imagenette2.tgz")

        print("Extracting...")
        tar = tarfile.open(path + "imagenette/imagenette2.tgz", "r:gz")
        tar.extractall(path + "imagenette/")
        tar.close()


def load_image(tar, name):
    file = tar.extractfile(name)
    image = matplotlib.image.imread(file, "JPEG")
    print("image", image)
    return image


def load(path=None, n_processes=6):
    """
    Parameters
    ----------
        path: str (optional)
            default ($DATASET_PATH), the path to look for the data and
            where the data will be downloaded if not present

    Returns
    -------

        train_images: array

        train_labels: array

        test_images: array

        test_labels: array

    """

    if path is None:
        path = os.environ["DATASET_PATH"]

    download(path)

    t0 = time.time()

    # Load train set
    train_images = list()
    train_labels = list()
    test_images = list()
    test_labels = list()

    tar = tarfile.open(path + "imagenette/imagenette2.tgz", "r:gz")

    train_names = [
        name for name in tar.getnames() if "JPEG" in name and "train" in name
    ]
    test_names = [
        name for name in tar.getnames() if "JPEG" in name and "train" not in name
    ]

    for name in tqdm(train_names, ascii=True, desc="Loading training set"):
        image = matplotlib.image.imread(path + "imagenette/" + name)
        if image.ndim == 2:
            image = image[:, :, None] * np.ones((1, 1, 3))
        train_images.append(image)
        train_labels.append(name.split("/")[2])

    for name in tqdm(test_names, ascii=True, desc="Loading test set"):
        image = matplotlib.image.imread(path + "imagenette/" + name)
        if image.ndim == 2:
            image = image[:, :, None] * np.ones((1, 1, 3))
        test_images.append(image)
        test_labels.append(name.split("/")[2])

    train_labels = (np.unique(train_labels) == np.array(train_labels)[:, None]).argmax(
        1
    )
    test_labels = (np.unique(test_labels) == np.array(test_labels)[:, None]).argmax(1)

    data = {
        "train_set/images": train_images,
        "train_set/labels": train_labels.astype("int32"),
        "test_set/images": test_images,
        "test_set/labels": test_labels.astype("int32"),
    }

    tar.close()

    print("Dataset imagenette loaded in{0:.2f}s.".format(time.time() - t0))

    return data



from .. import tensor as T
import numpy as np
from .ops_nn import softmax, log_softmax


def vae(x, x_hat, z_mu, z_logvar, mu, logvar, logvar_x=0.0, eps=1e-8):
    """N samples of dimension D to latent space in K dimension

    Parameters
    ----------

    x: array
        should be of shape (N, D)

    x_hat: array
        should be of shape (N, D)

    z_mu: array
        should be of shape (N, K), infered mean of variational Gaussian

    z_logvar: array
        should be of shape (N, K), infered log-variance of variational Gaussian

    mu: array
        should be of shape (K,), parameter (centroids)

    logvar: array
        should be of shape (K,), parameter (logvar of clusters)
        :param logvar_x:
        :param eps:

    """

    var = T.exp(logvar)

    # E_{q(z,c|x)}[log(p(x|z))]
    px_z = -0.5 * ((x - x_hat) ** 2 / T.exp(logvar_x) + logvar_x).sum(1)

    # - E_{q(z,c|x)}[log(q(z|x))] : entropy of normal
    h_z = 0.5 * z_logvar.sum(1)

    # E_{q(z,c|x)}[log(p(z|c)]
    ll_z = -0.5 * (logvar + z_var / var - 1 + (z_mu - mu) ** 2 / var).sum(-1)

    loss = -(px_z + ll_z + h_z)

    return loss


def vae_gmm(x, x_hat, z_mu, z_logvar, mu, logvar, logpi, logvar_x=0.0, eps=1e-8):
    """N samples of dimension D to latent space of C sluters in K dimension

    Parameters
    ----------

    x: array
        should be of shape (N, D)

    x_hat: array
        should be of shape (N, D)

    z_mu: array
        should be of shape (N, K), infered mean of variational Gaussian

    z_logvar: array
        should be of shape (N, K), infered log-variance of variational Gaussian

    mu: array
        should be of shape (C, K), parameter (centroids)

    logvar: array
        should be of shape (C, K), parameter (logvar of clusters)

    logpi: array
        should be of shape (C,), parameter (prior of clusters)
        :param logvar_x:
        :param eps:

    """

    var = T.exp(logvar)
    z_var = T.exp(z_logvar)

    # predict the log probability of clusters, shape will be (N, C)
    # and compute compute p(t|z) = p(z|t)p(t)/(\sum_t p(z|t)p(t))
    logprob = (
        logpi[:, None]
        - 0.5 * (T.log(2 * np.pi) + logvar)
        - (z_mu[:, None, :] - mu) ** 2 / (2 * var)
    ).sum(2)
    pt_z = T.softmax(logprob)

    # E_{q(z,c|x)}[log(p(x|z))]
    px_z = -0.5 * ((x - x_hat) ** 2 / T.exp(logvar_x) + logvar_x).sum(1)

    # - E_{q(z,c|x)}[log(q(c|x))] entropy of categorical
    h_c = -(pt_z * T.log_softmax(logprob)).sum(1)

    # - E_{q(z,c|x)}[log(q(z|x))] : entropy of normal
    h_z = 0.5 * z_logvar.sum(1)

    # E_{q(z,c|x)}[log(p(z|c)]
    ll_z = -0.5 * (
        pt_z
        * (
            logvar + z_var[:, None, :] / var - 1 + (z_mu[:, None, :] - mu) ** 2 / var
        ).sum(-1)
    ).sum(-1)

    # E_{q(z,c|x)}[log(p(c)]
    p_c = (pt_z * logpi).sum(1)

    loss = -(px_z + ll_z + p_c + h_c + h_z)

    return loss, px_z + p_c, pt_z


def vae_comp_gmm(x, x_hat, z_mu, z_logvar, mu, logvar, logpi, logvar_x=0.0, eps=1e-8):
    """N samples of dimension D to latent space of I pieces each of C sluters
    in K dimension

    Parameters
    ----------

    x: array
        should be of shape (N, D)

    x_hat: array
        should be of shape (N, D)

    z_mu: array
        should be of shape (N, I, K), infered mean of variational Gaussian

    z_logvar: array
        should be of shape (N, I, K), infered log-variance of variational Gaussian

    mu: array
        should be of shape (I, C, K), parameter (centroids)

    logvar: array
        should be of shape (I, C, K), parameter (logvar of clusters)

    logpi: array
        should be of shape (I, C), parameter (prior of clusters)
        :param logvar_x:
        :param eps:

    """

    var = T.exp(logvar)
    z_var = T.exp(z_logvar)

    # predict the log probability of clusters, shape will be (N, I, C)
    # and compute compute p(t_i|z_i) = p(z_i|t_i)p(t_i)/(\sum_t_i p(z_i|t_i)p(t_i))
    logprob = (
        logpi[:, :, None]
        - 0.5 * (T.log(2 * np.pi) + logvar)
        - (z_mu[:, :, None, :] - mu) ** 2 / (2 * var)
    ).sum(3)
    pt_z = softmax(logprob)

    # E_{q(z,c|x)}[log(p(x|z))]
    px_z = ((x - x_hat) ** 2).sum(1)

    # - E_{q(z,c|x)}[log(q(c|x))] entropy of categorical
    h_c = -(pt_z * log_softmax(logprob)).sum((1, 2))

    # - E_{q(z,c|x)}[log(q(z|x))] : entropy of normal
    h_z = 0.5 * z_logvar.sum((1, 2))

    # E_{q(z,c|x)}[log(p(z|c)]
    ll_z = -0.5 * (
        pt_z
        * (
            logvar
            + z_var[:, :, None, :] / var
            - 1
            + (z_mu[:, :, None, :] - mu) ** 2 / var
        ).sum(-1)
    ).sum((1, 2))

    # E_{q(z,c|x)}[log(p(c)]
    p_c = (pt_z * logpi).sum((1, 2))

    loss = -(px_z + ll_z + p_c + h_c + h_z)

    return loss


def FGMM_VAE(
    x,
    x_rec,
    x_logvar,
    z_logvar,
    q_mu,
    q_logvar,
    mu,
    q_loggamma,
    q_logeta,
    logpi,
    logpia,
    mode="bernoulli",
    eps=1e-5,
):
    """N samples of dimension D to latent space of dimension K with F factors of C clusters
    in K dimension

    Parameters
    ----------

    x: array
        (observation) should be of shape (N, D)

    x_rec: array
        (reconstruction, output of decoder) should be of shape (N, D)

    x_logvar: array
        (parameter) should be of shape (D,)

    z_logvar: array
        (parameter) should be of shape (K,)

    q_mu: array
        should be of shape (N, K), infered mean of variational Gaussian

    q_logvar: array
        should be of shape (N, K), infered log-variance of variational Gaussian

    mu: array
        (parameter) should be of shape (F, C, K), parameter (centroids)

    q_loggamma: array
        should be of shape (N, F, C), parameter (prior of clusters)

    logpi: array
        (parameter) should be of shape (F, C), parameter (prior of clusters)

    logpia: array
        (parameters) should ve of shape (F,)
        :param q_logeta:
        :param mode:
        :param eps:

    """

    q_var = T.exp(q_logvar)
    q_gamma = T.exp(q_loggamma)
    q_eta = T.exp(q_logeta)
    z_var = T.exp(z_logvar)
    x_var = T.exp(x_logvar)
    pi = T.exp(logpi)
    pia = T.exp(logpia)

    K = z_var.shape[0]
    D = x_var.shape[0]
    F = logpi.shape[0]
    log2pi = T.log(2 * np.pi)

    # reconstruction part (first expectation)
    E1 = -0.5 * (((x - x_rec) ** 2 / x_var).sum(1) + x_logvar.sum() + D * log2pi)

    E2_1 = -0.5 * (log2pi + z_logvar + (q_var + q_mu ** 2) / z_var).sum(1)

    E2_2 = T.einsum("nf,nfc,fck,nk->n", q_eta, q_gamma, mu, q_mu / z_var)

    E2_3 = -0.5 * (T.einsum("nf,nfc,fck->nk", q_eta, q_gamma, mu ** 2) / z_var).sum(1)

    if mode == "bernoulli":
        q_gammaeta = T.einsum("nf,nfc->nfc", q_eta, q_gamma)
        corr = T.einsum("fcd,nfc,abk,nab->nfa", mu / z_var, q_gammaeta, mu, q_gammaeta)
        E2_4 = -0.5 * T.sum(corr * (1 - T.eye(F)), (1, 2))
    else:
        E2_4 = 0.0
    E3 = (q_gamma * logpi).sum((1, 2))
    if mode == "bernoulli":
        E4 = (q_eta * logpia + (1 - q_eta) * T.log(1 - pia + eps)).sum(1)
    else:
        E4 = (q_eta * logpia).sum(1)

    # now on to the entropy
    H = (
        K * (log2pi + 1) / 2
        + 0.5 * q_logvar.sum(1)
        - (q_gamma * q_loggamma).sum((1, 2))
    )
    if mode == "bernoulli":
        Ha = -(q_eta * q_logeta + (1 - q_eta) * T.log(1 - q_eta + eps)).sum(1)
    else:
        Ha = -(q_eta * q_logeta).sum(1)
    return -(E1 + E2_1 + E2_2 + E2_3 + E2_4 + E3 + E4 + H + Ha)


def sparse_crossentropy_logits(p, q, weights=None):
    """Cross entropy loss given that :math:`p` is sparse and
    :math:`q` is the log-probability.

    The formal definition given that :math:`p` is now an
    index (of the Dirac) s.a. :math:`p\in \{1,\dots,D\}`
    and :math:`q` is unormalized (log-proba)
    is given by (for discrete variables, p sparse)

    .. math::
        \mathcal{L}(p,q)=-q_{p}+\log(\sum_{d=1}^D \exp(q_d))
    .. math::
        \mathcal{L}(p,q)=-q_{p}+LogSumExp(q)
    .. math::
        \mathcal{L}(p,q)=-q_{p}+LogSumExp(q-\max_{d}q_d)

    or by (non p sparse)

    .. math::
        \mathcal{L}(p,q)=-\sum_{d=1}^Dp_{d}q_{d}+\log(\sum_{d=1}^D \exp(q_d))
    .. math::
        \mathcal{L}(p,q)=-\sum_{d=1}^Dp_{d}q_{d}+LogSumExp(q)
    .. math::
        \mathcal{L}(p,q)=-\sum_{d=1}^Dp_{d}q_{d}+LogSumExp(q-\max_{d}q_d)


    with :math:`p` the class index and :math:`q` the predicted one
    (output of the network). This class takes two non sparse
    vectors which should be nonnegative and sum to one.
    """
    # the linear part of the loss
    linear = T.take_along_axis(q, p[:, None], 1).squeeze()
    logsumexp = T.logsumexp(q, 1)
    if weights is not None:
        return weights * (-linear + logsumexp)
    else:
        return logsumexp - linear


def crossentropy_logits(p, q, p_sparse=True):
    """see sparse cross entropy"""
    linear = (p * q).sum(1)
    logsumexp = T.logsumexp(q, 1)
    return -linear + logsumexp


def sigmoid_crossentropy_logits(labels, logits):
    return -logits * labels + T.log1p(T.exp(logits))


def multiclass_hinge_loss(predictions, targets, delta=1):
    """Computes the multi-class hinge loss between predictions and targets.
    .. math:: L_i = \\max_{j \\not = t_i} (0, p_j - p_{t_i} + \\delta)
    Parameters
    ----------
    predictions : Theano 2D tensor
        Predictions in (0, 1), such as softmax output of a neural network,
        with data points in rows and class probabilities in columns.
    targets : Theano 2D tensor or 1D tensor
        Either a vector of int giving the correct class index per data point
        or a 2D tensor of one-hot encoding of the correct class in the same
        layout as predictions (non-binary targets in [0, 1] do not work!)
    delta : scalar, default 1
        The hinge loss margin
    Returns
    -------
    Theano 1D tensor
        An expression for the item-wise multi-class hinge loss
    Notes
    -----
    This is an alternative to the categorical cross-entropy loss for
    multi-class classification problems
    """
    num_cls = predictions.shape[1]
    if targets.ndim == predictions.ndim - 1:
        targets = T.to_one_hot(targets, num_cls)
    elif targets.ndim != predictions.ndim:
        raise TypeError("rank mismatch between targets and predictions")
    corrects = predictions[targets.nonzero()]
    rest = predictions[(1 - targets).nonzero()].reshape((-1, num_cls - 1))
    rest = rest.max(axis=1)
    return T.activations.relu(rest - corrects + delta)


def squared_difference(targets, predictions):
    return (targets - predictions) ** 2


def accuracy(targets, predictions):
    if predictions.ndim == 2:
        accu = T.cast(T.equal(targets, predictions.argmax(1)), "float32")
    else:
        accu = T.cast(T.equal(targets, predictions), "float32")
    return accu.mean()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import time


class Identity:
    def __init__(self, eps=0.0001, name=""):
        self.name = name + "preprocessing(identity)"

    def fit(self, x, *args, **kwargs):
        return self

    def fit_transform(self, x, *args, **kwargs):
        return x

    def transform(self, x, *args, **kwargs):
        return x


class Standardize:
    def __init__(self, eps=0.000001, axis=[0], name=""):
        self.name = name + "preprocessing(standardize,eps=" + str(eps) + ")"
        self.eps = eps
        self.axis = axis

    def fit(self, x, **kwargs):
        print(self.name + " fitting...")
        t = time.time()
        self.mean = x.mean(axis=tuple(self.axis), keepdims=True)
        self.std = x.std(axis=tuple(self.axis), keepdims=True) + self.eps
        print(self.name + " done in {0:.2f} s.".format(time.time() - t))
        return self

    def transform(self, x, inplace=False, **kwargs):
        if inplace:
            x -= self.mean
            x /= self.std
        else:
            return (x - self.mean) / self.std

    def fit_transform(self, x, inplace=False, **kwargs):
        self.fit(x)
        return self.transform(x, inplace)


class ZCAWhitening:
    def __init__(self, eps=0.0001, name=""):
        self.name = name + "preprocessing(zcawhitening,eps=" + str(eps) + ")"
        self.eps = eps

    def fit(self, x):
        print(self.name + " fitting ...")
        t = time.time()
        flatx = np.reshape(x, (x.shape[0], -1))
        self.mean = flatx.mean(0, keepdims=True)
        self.S, self.U = _spectral_decomposition(flatx - self.mean, self.eps)
        print(self.name + " done in {0:.2f} s.".format(time.time() - t))
        return self

    def transform(self, x):
        flatx = np.reshape(x, (x.shape[0], -1)) - self.mean
        return _zca_whitening(flatx, self.U, self.S).reshape(x.shape)

    def fit_transform(self, x):
        self.fit(x)
        return self.transform(x)


def _spectral_decomposition(flatx, eps):
    U, S, V = np.linalg.svd(flatx, full_matrices=False)
    S = np.diag(1.0 / np.sqrt(S + eps))
    return S, V


def _zca_whitening(flatx, U, S):
    M = np.dot(np.dot(U.T, S), U)
    return np.dot(M, flatx.T).T



import sys
from .base import jax_wrap
import jax.numpy as jnp

# Add the fft functions

module = sys.modules[__name__]
names = [
    "fft",
    "ifft",
    "fft2",
    "ifft2",
    "fftn",
    "ifftn",
    "rfft",
    "irfft",
    "rfft2",
    "irfft2",
    "rfftn",
    "irfftn",
    "fftfreq",
    "rfftfreq",
    "ifftshift",
    "fftshift",
]

for name in names:
    module.__dict__.update({name: jax_wrap(jnp.fft.__dict__[name])})



import os
import pickle
import tarfile
import time
import urllib.request

import numpy as np


DESCRIPTION = """Image classification.

The `CIFAR-100 < https: // www.cs.toronto.edu/~kriz/cifar.html >`_ dataset is
just like the CIFAR-10, except it has 100 classes containing 600 images
each. There are 500 training images and 100 testing images per class.
The 100 classes in the CIFAR-100 are grouped into 20 superclasses. Each
image comes with a "fine" label(the class to which it belongs) and a
"coarse" label(the superclass to which it belongs).
"""

labels_list = [
    "apple",
    "aquarium_fish",
    "baby",
    "bear",
    "beaver",
    "bed",
    "bee",
    "beetle",
    "bicycle",
    "bottle",
    "bowl",
    "boy",
    "bridge",
    "bus",
    "butterfly",
    "camel",
    "can",
    "castle",
    "caterpillar",
    "cattle",
    "chair",
    "chimpanzee",
    "clock",
    "cloud",
    "cockroach",
    "couch",
    "crab",
    "crocodile",
    "cup",
    "dinosaur",
    "dolphin",
    "elephant",
    "flatfish",
    "forest",
    "fox",
    "girl",
    "hamster",
    "house",
    "kangaroo",
    "keyboard",
    "lamp",
    "lawn_mower",
    "leopard",
    "lion",
    "lizard",
    "lobster",
    "man",
    "maple_tree",
    "motorcycle",
    "mountain",
    "mouse",
    "mushroom",
    "oak_tree",
    "orange",
    "orchid",
    "otter",
    "palm_tree",
    "pear",
    "pickup_truck",
    "pine_tree",
    "plain",
    "plate",
    "poppy",
    "porcupine",
    "possum",
    "rabbit",
    "raccoon",
    "ray",
    "road",
    "rocket",
    "rose",
    "sea",
    "seal",
    "shark",
    "shrew",
    "skunk",
    "skyscraper",
    "snail",
    "snake",
    "spider",
    "squirrel",
    "streetcar",
    "sunflower",
    "sweet_pepper",
    "table",
    "tank",
    "telephone",
    "television",
    "tiger",
    "tractor",
    "train",
    "trout",
    "tulip",
    "turtle",
    "wardrobe",
    "whale",
    "willow_tree",
    "wolf",
    "woman",
    "worm",
]


def download(path):
    """
    Download the CIFAR100 dataset and store the result into the given
    path

    Parameters
    ----------

        path: str
            the path where the downloaded files will be stored. If the
            directory does not exist, it is created.
    """

    print("Loading cifar100")
    t0 = time.time()

    if not os.path.isdir(path + "cifar100"):
        print("\tCreating cifar100 Directory")
        os.mkdir(path + "cifar100")

    if not os.path.exists(path + "cifar100/cifar100.tar.gz"):
        url = "https://www.cs.toronto.edu/~kriz/cifar-100-python.tar.gz"
        #        with DownloadProgressBar(
        #            unit="B", unit_scale=True, miniters=1, desc="Downloading dataset"
        #        ) as t:
        urllib.request.urlretrieve(url, path + "cifar100/cifar100.tar.gz")


def load(path=None):

    t0 = time.time()
    if path is None:
        path = os.environ["DATASET_PATH"]
    download(path)

    # Loading the file
    tar = tarfile.open(path + "cifar100/cifar100.tar.gz", "r:gz")

    # Loading training set
    f = tar.extractfile("cifar-100-python/train").read()
    data = pickle.loads(f, encoding="latin1")
    train_images = data["data"].reshape((-1, 3, 32, 32)).astype("float32")
    train_fine = np.array(data["fine_labels"])
    train_coarse = np.array(data["coarse_labels"])

    # Loading test set
    f = tar.extractfile("cifar-100-python/test").read()
    data = pickle.loads(f, encoding="latin1")
    test_images = data["data"].reshape((-1, 3, 32, 32)).astype("float32")
    test_fine = np.array(data["fine_labels"])
    test_coarse = np.array(data["coarse_labels"])

    data = {
        "train_set/images": train_images,
        "train_set/labels": train_fine,
        "train_set/coarse_labels": train_coarse,
        "test_set/images": test_images,
        "test_set/labels": test_fine,
        "test_set/coarse_labels": test_coarse,
    }

    print("Dataset cifar100 loaded in {0:.2f}s.".format(time.time() - t0))
    return data



import scipy.io as sio
import os
import pickle, gzip
import urllib.request
import numpy as np
import time


class svhn:
    """Street number classification.

    The `SVHN <http://ufldl.stanford.edu/housenumbers/>`_
    dataset is a real-world 
    image dataset for developing machine learning and object 
    recognition algorithms with minimal requirement on data 
    preprocessing and formatting. It can be seen as similar in flavor 
    to MNIST (e.g., the images are of small cropped digits), but 
    incorporates an order of magnitude more labeled data (over 600,000 
    digit images) and comes from a significantly harder, unsolved, 
    real world problem (recognizing digits and numbers in natural 
    scene images). SVHN is obtained from house numbers in Google 
    Street View images. 
    """

    def download(path):
        """
        Download the SVHN dataset and store the result into the given
        path

        Parameters
        ----------

            path: str
                the path where the downloaded files will be stored. If the
                directory does not exist, it is created.
        """

        # Load the dataset (download if necessary) and set
        # the class attributess.
        print("Loading svhn")

        t0 = time.time()

        if not os.path.isdir(path + "svhn"):
            os.mkdir(path + "svhn")
            print("\tCreating svhn Directory")

        if not os.path.exists(path + "svhn/train_32x32.mat"):
            url = "http://ufldl.stanford.edu/housenumbers/train_32x32.mat"
            urllib.request.urlretrieve(url, path + "svhn/train_32x32.mat")

        if not os.path.exists(path + "svhn/test_32x32.mat"):
            url = "http://ufldl.stanford.edu/housenumbers/test_32x32.mat"
            urllib.request.urlretrieve(url, path + "svhn/test_32x32.mat")

    def load(path=None):
        """
        Parameters
        ----------
            path: str (optional)
                default $DATASET_PATH, the path to look for the data and
                where the data will be downloaded if not present

        Returns
        -------

            train_images: array

            train_labels: array

            test_images: array

            test_labels: array


        """
        if path is None:
            path = os.environ["DATASET_PATH"]

        svhn.download(path)

        # Load the dataset (download if necessary) and set
        # the class attributess.
        print("Loading svhn")

        t0 = time.time()

        # Train set
        data = sio.loadmat(path + "svhn/train_32x32.mat")
        train_images = data["X"].transpose([3, 2, 0, 1])
        train_labels = np.squeeze(data["y"]) - 1

        # Test set
        data = sio.loadmat(path + "svhn/test_32x32.mat")
        test_images = data["X"].transpose([3, 2, 0, 1])
        test_labels = np.squeeze(data["y"]) - 1

        print("Dataset svhn loaded in", "{0:.2f}".format(time.time() - t0), "s.")
        return train_images, train_labels, test_images, test_labels



import os
import gzip
import urllib.request
import numpy as np
import time
import zipfile
import io
from scipy.io.wavfile import read as wav_read
from tqdm import tqdm


class warblr:
    """Binary audio classification, presence or absence of a bird.

    `Warblr <http://machine-listening.eecs.qmul.ac.uk/bird-audio-detection-challenge/#downloads>`_ 
    comes from a UK bird-sound crowdsourcing 
    research spinout called Warblr. From this initiative we have 
    10,000 ten-second smartphone audio recordings from around the UK. 
    The audio totals around 44 hours duration. The audio will be 
    published by Warblr under a Creative Commons licence. The audio 
    covers a wide distribution of UK locations and environments, and 
    includes weather noise, traffic noise, human speech and even human 
    bird imitations. It is directly representative of the data that is 
    collected from a mobile crowdsourcing initiative.
    """

    def download(path):
        """
        Download the data
        """

        # Load the dataset (download if necessary) and set
        # the class attributes.
        print("Loading warblr")
        t = time.time()
        if not os.path.isdir(path + "warblr"):
            print("\tCreating Directory")
            os.mkdir(path + "warblr")

        if not os.path.exists(path + "warblr/warblrb10k_public_wav.zip"):
            url = "https://archive.org/download/warblrb10k_public/warblrb10k_public_wav.zip"
            urllib.request.urlretrieve(url, path + "warblr/warblrb10k_public_wav.zip")

        if not os.path.exists(path + "warblr/warblrb10k_public_metadata.csv"):
            url = "https://ndownloader.figshare.com/files/6035817"
            urllib.request.urlretrieve(
                url, path + "warblr/warblrb10k_public_metadata.csv"
            )

    def load(path=None):
        """
        Load the data given a path
        """

        if path is None:
            path = os.environ["DATASET_PATH"]

        warblr.download(path)

        # Load the dataset (download if necessary) and set
        # the class attributes.
        print("Loading warblr")
        t = time.time()

        # Loading Labels
        labels = np.loadtxt(
            path + "warblr/warblrb10k_public_metadata.csv",
            delimiter=",",
            skiprows=1,
            dtype="str",
        )
        # Loading the files
        f = zipfile.ZipFile(path + "warblr/warblrb10k_public_wav.zip")
        N = labels.shape[0]
        wavs = list()
        for i, files_ in tqdm(enumerate(labels), ascii=True):
            wavfile = f.read("wav/" + files_[0] + ".wav")
            byt = io.BytesIO(wavfile)
            wavs.append(np.expand_dims(wav_read(byt)[1].astype("float32"), 0))
        labels = labels[:, 1].astype("int32")

        print("Dataset warblr loaded in", "{0:.2f}".format(time.time() - t), "s.")
        return wavs, labels



import jax.lax as jla

from .base import jax_wrap, symjax_to_jax_fn

cond = jax_wrap(jla.cond)
fori_loop = jax_wrap(jla.fori_loop)
while_loop = jax_wrap(jla.while_loop)


@jax_wrap
def _scan(f, init, sequences, non_sequences=None, length=None, reverse=False):
    # get the fully jaxed function
    truef = symjax_to_jax_fn(f)

    # now create a dummy function that only takes as input the sequences
    if non_sequences is None:

        def finalf(a, args):
            return truef(a, *args)

    else:

        def finalf(a, args):
            return truef(a, *args, *non_sequences)

    return jla.scan(finalf, init, sequences, length=length, reverse=reverse)


@jax_wrap
def _while_loop(
    cond_fun, body_fun, sequences, non_sequences_cond=None, non_sequences_body=None
):

    # get the fully jaxed function
    truecond = symjax_to_jax_fn(cond_fun)
    truebody = symjax_to_jax_fn(body_fun)

    # now create a dummy function that only takes as input the sequences
    if non_sequences_cond is None:
        finalcond = truecond
    else:

        def finalcond(args):
            return truecond(args, *non_sequences_cond)

    if non_sequences_body is None:
        finalbody = truebody
    else:

        def finalbody(args):
            return truebody(args, *non_sequences_body)

    return jla.while_loop(finalcond, finalbody, sequences)


def map(f, sequences, non_sequences=None):
    """Map a function over leading array axes.

    Like Python's builtin map, except inputs and outputs are in the form of
    stacked arrays. Consider using the ``jax.vmap`` transform instead, unless you
    need to apply a function element by element for reduced memory usage or
    heterogeneous computation with other control flow primitives.

    When ``xs`` is an array type, the semantics of ``map`` are given by this
    Python implementation::

      def map(f, xs):
        return np.stack([f(x) for x in xs])

    Like ``scan``, ``map`` is implemented in terms of JAX primitives so many of
    the same advantages over a Python loop apply: ``xs`` may be an arbitrary
    nested pytree type, and the mapped computation is compiled only once.

    Args:
      f: a Python function to apply element-wise over the first axis or axes of
        ``sequences``.
      sequences: list of values over which to map along the leading axis.
      non_sequences: list of values passed the same at each call

    Returns:
      Mapped values.

    Example:
        example of creating a diagonal matrix:
        .. doctest::

           >>> import symjax.tensor as T
           >>> import symjax
           >>> x = T.ones(3)
           >>> y = T.zeros(3)
           >>> w = T.arange(3)
           >>> out = T.map(lambda x, i, w: T.index_update(w, i, x), sequences=[x, w], non_sequences=[y])
           >>> f = symjax.function(outputs=out)
           >>> f()
           array([[1., 0., 0.],
                  [0., 1., 0.],
                  [0., 0., 1.]], dtype=float32)
    """

    g = lambda _, *args: (1, f(*args))

    if type(non_sequences) == list:
        non_sequences = tuple(non_sequences)
    if type(sequences) == list:
        sequences = tuple(sequences)
    ys = scan(g, 0, sequences, non_sequences=non_sequences)[1]

    return ys


def scan(f, init, sequences, non_sequences=None, length=None, reverse=False):
    """Scan a function over leading array axes while carrying along state.

    The type signature in brief is

    .. code-block:: haskell

      scan :: (c -> a -> (c, b)) -> c -> [a] -> (c, [b])

    where we use [t] here to denote the type t with an additional leading axis.
    That is, if t is an array type then [t] represents the type with an additional
    leading axis, and if t is a pytree (container) type with array leaves then [t]
    represents the type with the same pytree structure and corresponding leaves
    each with an additional leading axis.

    When ``a`` is an array type or None, and ``b`` is an array type, the semantics
    of ``scan`` are given roughly by this Python implementation::

      def scan(f, init, xs, length=None):
        if xs is None:
          xs = [None] * length
        carry = init
        ys = []
        for x in xs:
          carry, y = f(carry, x)
          ys.append(y)
        return carry, np.stack(ys)

    Unlike that Python version, both ``a`` and ``b`` may be arbitrary pytree
    types, and so multiple arrays can be scanned over at once and produce multiple
    output arrays. (None is actually an empty pytree.)

    Also unlike that Python version, ``scan`` is a JAX primitive and is lowered to
    a single XLA While HLO. That makes it useful for reducing compilation times
    for jit-compiled functions, since native Python loop constructs in an ``@jit``
    function are unrolled, leading to large XLA computations.

    Finally, the loop-carried value ``carry`` must hold a fixed shape and dtype
    across all iterations (and not just be consistent up to NumPy rank/shape
    broadcasting and dtype promotion rules, for example). In other words, the type
    ``c`` in the type signature above represents an array with a fixed shape and
    dtype (or a nested tuple/list/dict container data structure with a fixed
    structure and arrays with fixed shape and dtype at the leaves).

    Args:
      f: a Python function to be scanned of type ``c -> a -> (c, b)``, meaning
        that ``f`` accepts two arguments where the first is a value of the loop
        carry and the second is a slice of ``xs`` along its leading axis, and that
        ``f`` returns a pair where the first element represents a new value for
        the loop carry and the second represents a slice of the output.
      init: an initial loop carry value of type ``c``, which can be a scalar,
        array, or any pytree (nested Python tuple/list/dict) thereof, representing
        the initial loop carry value. This value must have the same structure as
        the first element of the pair returned by ``f``.
      sequences: list of values over which to scan along the leading axis,
        where ``[a]`` can be an array or any pytree (nested Python
        tuple/list/dict) thereof with consistent leading axis sizes.
      non_sequences: (optional) list of values that do not depend of the
        iterations but are used in the function
      length: optional integer specifying the number of loop iterations, which
        must agree with the sizes of leading axes of the arrays in ``sequences`` (but can
        be used to perform scans where no input ``xs`` are needed).
      reverse: optional boolean specifying whether to run the scan iteration
        forward (the default) or in reverse, equivalent to reversing the leading
        axes of the arrays in both ``xs`` and in ``ys``.

    Returns:
      A pair of type ``(c, [b])`` where the first element represents the final
      loop carry value and the second element represents the stacked outputs of
      the second output of ``f`` when scanned over the leading axis of the inputs.

    Example:
        .. doctest::

           >>> import symjax.tensor as T
           >>> import symjax
           >>> indices = T.Placeholder((3,), 'int32')
           >>> w = T.arange(10)
           >>> out = T.scan(lambda acc, i, w, : (acc + w[i],i), T.zeros(1), sequences=[indices], non_sequences=[w])
           >>> # out first element is the carry, the second is the stacked second outputs of the function
           >>> f = symjax.function(indices, outputs=out)
           >>> f([0,2,3])
           [array([5.], dtype=float32), array([0, 2, 3], dtype=int32)]
           >>> f([1, 1, 1])
           [array([3.], dtype=float32), array([1, 1, 1], dtype=int32)]
    """
    if type(non_sequences) == list:
        non_sequences = tuple(non_sequences)

    if type(sequences) == list:
        sequences = tuple(sequences)

    return _scan(f, init, sequences, non_sequences, length, reverse)


def while_loop(
    cond_fun, body_fun, sequences, non_sequences_cond=None, non_sequences_body=None
):
    """Call ``body_fun`` repeatedly in a loop while ``cond_fun`` is True.

     The type signature in brief is

     .. code-block:: haskell

       while_loop :: (a -> Bool) -> (a -> a) -> a -> a

     The semantics of ``while_loop`` are given by this Python implementation::

       def while_loop(cond_fun, body_fun, init_val):
         val = init_val
         while cond_fun(val):
           val = body_fun(val)
         return val

     Unlike that Python version, ``while_loop`` is a JAX primitive and is lowered
     to a single XLA While HLO. That makes it useful for reducing compilation times
     for jit-compiled functions, since native Python loop constructs in an ``@jit``
     function are unrolled, leading to large XLA computations.

     Also unlike the Python analogue, the loop-carried value ``val`` must hold a
     fixed shape and dtype across all iterations (and not just be consistent up to
     NumPy rank/shape broadcasting and dtype promotion rules, for example). In
     other words, the type ``a`` in the type signature above represents an array
     with a fixed shape and dtype (or a nested tuple/list/dict container data
     structure with a fixed structure and arrays with fixed shape and dtype at the
     leaves).

     Another difference from using Python-native loop constructs is that
     ``while_loop`` is not reverse-mode differentiable because XLA computations
     require static bounds on memory requirements.

     Args:
       cond_fun: function of type ``a -> Bool``.
       body_fun: function of type ``a -> a``.
       init_val: value of type ``a``, a type that can be a scalar, array, or any
         pytree (nested Python tuple/list/dict) thereof, representing the initial
         loop carry value.

     Returns:
       The output from the final iteration of body_fun, of type ``a``.
    """

    if type(non_sequences_cond) == list:
        non_sequences_cond = tuple(non_sequences_cond)
    if type(non_sequences_body) == list:
        non_sequences_body = tuple(non_sequences_body)
    if type(sequences) == list:
        sequences = tuple(sequences)

    return _while_loop(
        cond_fun, body_fun, sequences, non_sequences_cond, non_sequences_body
    )



import inspect
import sys

import jax.numpy.linalg as jnpl

from .base import jax_wrap

NAMES = [c[0] for c in inspect.getmembers(jnpl, inspect.isfunction)]
module = sys.modules[__name__]
for name in NAMES:
    module.__dict__.update({name: jax_wrap(jnpl.__dict__[name])})



import os
import pickle
import tarfile
import time
import urllib.request

import numpy as np
from tqdm import tqdm


__DESCRIPTION__ = """Image classification.
The `CIFAR-10 < https: // www.cs.toronto.edu/~kriz/cifar.html >`_ dataset
was collected by Alex Krizhevsky, Vinod Nair, and Geoffrey
Hinton. It consists of 60000 32x32 colour images in 10 classes, with
6000 images per class. There are 50000 training images and 10000 test images.
The dataset is divided into five training batches and one test batch,
each with 10000 images. The test batch contains exactly 1000 randomly
selected images from each class. The training batches contain the
remaining images in random order, but some training batches may
contain more images from one class than another. Between them, the
training batches contain exactly 5000 images from each class.
"""

label_to_name = {
    0: "airplane",
    1: "automobile",
    2: "bird",
    3: "cat",
    4: "deer",
    5: "dog",
    6: "frog",
    7: "horse",
    8: "sheep",
    9: "truck",
}


def download(path):
    """
    Download the MNIST dataset and store the result into the given
    path

    Parameters
    ----------

        path: str
            the path where the downloaded files will be stored. If the
            directory does not exist, it is created.
    """

    t0 = time.time()

    # Check if directory exists
    if not os.path.isdir(path + "cifar10"):
        print("\tCreating cifar10 Directory")
        os.mkdir(path + "cifar10")

    # Check if file exists
    if not os.path.exists(path + "cifar10/cifar10.tar.gz"):
        print("\tDownloading cifar10")
        url = "https://www.cs.toronto.edu/~kriz/cifar-10-python.tar.gz"
        urllib.request.urlretrieve(url, path + "cifar10/cifar10.tar.gz")


def load(path=None):
    """
    Parameters
    ----------
        path: str (optional)
            default ($DATASET_PATH), the path to look for the data and
            where the data will be downloaded if not present

    Returns
    -------

        train_images: array

        train_labels: array

        test_images: array

        test_labels: array

    """

    if path is None:
        path = os.environ["DATASET_PATH"]

    download(path)

    t0 = time.time()

    tar = tarfile.open(path + "cifar10/cifar10.tar.gz", "r:gz")

    # Load train set
    train_images = list()
    train_labels = list()
    for k in tqdm(range(1, 6), desc="Loading cifar10", ascii=True):
        f = tar.extractfile("cifar-10-batches-py/data_batch_" + str(k)).read()
        data_dic = pickle.loads(f, encoding="latin1")
        train_images.append(data_dic["data"].reshape((-1, 3, 32, 32)))
        train_labels.append(data_dic["labels"])
    train_images = np.concatenate(train_images, 0).astype("float32")
    train_labels = np.concatenate(train_labels, 0).astype("int32")

    # Load test set
    f = tar.extractfile("cifar-10-batches-py/test_batch").read()
    data_dic = pickle.loads(f, encoding="latin1")
    test_images = data_dic["data"].reshape((-1, 3, 32, 32)).astype("float32")
    test_labels = np.array(data_dic["labels"]).astype("int32")

    data = {
        "train_set/images": train_images,
        "train_set/labels": train_labels,
        "test_set/images": test_images,
        "test_set/labels": test_labels,
    }

    print("Dataset cifar10 loaded in{0:.2f}s.".format(time.time() - t0))

    return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# flake8: noqa

from . import probabilities
from . import nn
from . import tensor
from . import data
from .base import *
from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions


_graphs = [Graph("default")]

__all__ = ["data", "tensor", "nn", "probabilities"]
